// defrag.c

#include <stdio.h>      // sprintf
//#include <io.h> // for win9x
//#include <winioctl.h>

#include "globals.h"
#include "resource.h"
#include "statbar.h"
#include "defrag.h"
#include "list.h"

// Size of the buffer we read file mapping information into.
// The buffer is big enough to hold the 16 bytes that
// come back at the head of the buffer (the number of entries
// and the starting virtual cluster), as well as 512 pairs
// of [virtual cluster, logical cluster] pairs.
#define FILEMAPSIZE (512+2)

// Size of the bitmap buffer we pass in. Its large enough to
// hold information for the 16-byte header that's returned
// plus the indicated number of bytes, each of which has 8 bits (imagine that!)
#define BITMAPBYTES 4096
#define BITMAPSIZE  (BITMAPBYTES+2*sizeof(ULONGLONG))

// Buffer to read file mapping information into
ULONGLONG FileMap[ FILEMAPSIZE ];

// Buffer thats passed to bitmap function
BYTE      BitMap[ BITMAPSIZE ];

// Bit shifting array for efficient processing of the bitmap
BYTE      BitShift[] = { 1, 2, 4, 8, 16, 32, 64, 128 };

// functions in this file
ULONGLONG     GetFreeClusters    (ULONGLONG MinLcn, ULONGLONG MinLen,
ULONGLONG     *beginLcn, ULONGLONG *endLcn, BOOL MatchSpace);
void          DumpBitmap         (void);
struct        DICT * DumpFile           (char *Path, char *filename, __int64 Bytes);

//void        MoveClusters(PMOVEFILE_DESCRIPTOR);
void          MoveClusters(PMOVE_FILE_DATA); // win32
ULONGLONG     MovePartOfFile(HANDLE sourceFile, int fragNo, ULONGLONG startVcn,
                             ULONGLONG OldLcn, ULONGLONG NewLcn, ULONGLONG fClusters);
ULONGLONG     MoveFileNormal(char *filename, ULONGLONG Lcn,
                             ULONGLONG     NewLcn, ULONGLONG Clusters);
ULONGLONG     MoveFileFragmented (char *filename, ULONGLONG *sLcn, ULONGLONG sLen);

long int progress=0, lastprogress=0;
int subDirLevel;

// OutLine0-3 TextOut at line 0-3 over the statusbar
void OutLine0(char *buff) {
    HDC hdc;
    SelectObject(memDC, hpBG);
    SelectObject(memDC, hbBG);
    Rectangle(memDC, 0, imgHeight-(dist*4), imgWidth, imgHeight-(dist*3+1));
    TextOut(memDC, 10, imgHeight-(dist*4), buff, strlen(buff));
    hdc = GetDC(ghWnd);
    BitBlt(hdc, 0, imgHeight-(dist*4), imgWidth, imgHeight-(dist*3+1), memDC, 0, imgHeight-(dist*4), SRCCOPY);
    ReleaseDC(ghWnd, hdc);
}

void OutLine1(char *buff) {
    HDC hdc;
    SelectObject(memDC, hpBG);
    SelectObject(memDC, hbBG);
    Rectangle(memDC, 0, imgHeight-(dist*3), imgWidth, imgHeight-(dist*2+1));
    TextOut(memDC, 10, imgHeight-(dist*3), buff, strlen(buff));
    hdc = GetDC(ghWnd);
    BitBlt(hdc, 0, imgHeight-(dist*3), imgWidth, imgHeight-(dist*2+1), memDC, 0, imgHeight-(dist*3), SRCCOPY);
    ReleaseDC(ghWnd, hdc);
}

void OutLine2(char *buff) {
    HDC hdc;
    SelectObject(memDC, hpBG);
    SelectObject(memDC, hbBG);
    Rectangle(memDC, 0, imgHeight-(dist*2), imgWidth, imgHeight-(dist+6));
    TextOut(memDC, 10, imgHeight-(dist*2), buff, strlen(buff));
    hdc = GetDC(ghWnd);
    BitBlt(hdc, 0, imgHeight-(dist*2), imgWidth, imgHeight-(dist+6), memDC, 0, imgHeight-(dist*2), SRCCOPY);
    ReleaseDC(ghWnd, hdc);
    strcpy(LastOutLine2, buff);
}

void OutLine3(char *buff) {
    HDC hdc;
    SelectObject(memDC, hpBG);
    SelectObject(memDC, hbBG);
    Rectangle(memDC, 10, imgHeight-(dist),imgWidth,imgHeight-6);
    TextOut(memDC, 10, imgHeight-(dist), buff, strlen(buff));
    hdc = GetDC(ghWnd);
    BitBlt(hdc, 0, imgHeight-(dist), imgWidth, imgHeight-6, memDC, 0, imgHeight-(dist), SRCCOPY);
    ReleaseDC(ghWnd, hdc);
    strcpy(LastOutLine3, buff);
}

// HandleNtError
// Translates an NTDLL error code into its text equivalent
void HandleNtError( NTSTATUS Status, char *zusatz) {
    char buff[100];
    switch( Status ) {
        case STATUS_SUCCESS:
        case STATUS_END_OF_FILE: return; // out of here
        case STATUS_DEVICE_BUSY: wsprintf(buff, "DEVICE_BUSY"); break;
        case STATUS_INVALID_PARAMETER: wsprintf(buff, "INVALID_PARAMETER"); break;
        case STATUS_INVALID_FILE_FOR_SECTION: wsprintf(buff, "INVALID_FILE_FOR_SECTION"); break;
        case STATUS_BUFFER_TOO_SMALL: wsprintf(buff, "BUFFER_TOO_SMALL"); break;
        case STATUS_ALREADY_COMMITTED: wsprintf(buff, "ALREADY_COMMITTED"); break;
        case STATUS_INVALID_DEVICE_REQUEST: wsprintf(buff, "INVALID_DEVICE_REQUEST"); break;
        case STATUS_INVALID_HANDLE: wsprintf(buff, "INVALID_HANDLE"); break;
        default: wsprintf(buff, "0x%08x", Status ); break;
    }
    lstrcat(buff, " ");
    lstrcat(buff, zusatz);
    //OutLine0(buff);
    SetLastError(0);
    //MessageBoxA(ghWnd, lpMsgBuf, (LPCTSTR) "Fehler", MB_OK);
}

// Translates a Win32 error into a text equivalent
void HandleWin32Error(DWORD ErrorCode, char *zusatz) {
    char buff[255];
    if (ErrorCode==0) return;
    FormatMessage( FORMAT_MESSAGE_ARGUMENT_ARRAY | FORMAT_MESSAGE_FROM_SYSTEM,
    NULL, ErrorCode,
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    (LPTSTR)buff, 255, NULL );
    lstrcat((LPTSTR)buff, (LPCTSTR)" ");
    lstrcat((LPTSTR)buff, (LPCTSTR)zusatz);
    OutLine1(buff);
    SetLastError(0);
    //MessageBox(ghWnd, (LPCSTR)buff, (LPCTSTR)"Info", MB_OK);
}

// ------------------------------------------------------------
void DrawNoBlockBound(int x, int y) {
    HDC hdc = GetDC(ghWnd);
    SelectObject(hdc, hpBG);
    SelectObject(hdc, GetStockObject(NULL_BRUSH));
    Rectangle(hdc, x, y, x+blocksize+1, y+blocksize+1);
    ReleaseDC(ghWnd, hdc);
    
    SelectObject(memDC, hpBG);
    SelectObject(memDC, GetStockObject(NULL_BRUSH));
    Rectangle(memDC, x, y, x+blocksize+1, y+blocksize+1);
}

void DrawBlockBound(int x, int y) {
    HDC hdc = GetDC(ghWnd);
    SelectObject(hdc, hpBlack);
    SelectObject(hdc, GetStockObject(NULL_BRUSH));
    Rectangle(hdc, x, y, x+blocksize+1, y+blocksize+1);
    ReleaseDC(ghWnd, hdc);
    
    SelectObject(memDC, hpBlack);
    SelectObject(memDC, GetStockObject(NULL_BRUSH));
    Rectangle(memDC, x, y, x+blocksize+1, y+blocksize+1);
}

void DrawBlockBounds(ULONGLONG Cstart, ULONGLONG Cend, BOOL clearmark) {
    ULONGLONG ui=0;
    int x=0, y=0;
    
    if (Cstart > cDiskClusters && Cend > cDiskClusters) return;
    if (Cend   > cDiskClusters) Cend   = cDiskClusters;
    if (Cstart > cDiskClusters) Cstart = cDiskClusters;
    
    while (                          y/blocksize*clustersPL <= Cstart )
    y += blocksize;
    y -= blocksize;
    
    while ( x/blocksize*clustersPB + y/blocksize*clustersPL <= Cstart )
    x += blocksize;
    x -= blocksize;
    
    for (ui=Cstart; ui<Cend; ui+=clustersPB) {
        if (!clearmark)
        DrawBlockBound(x, y);
        else
        DrawNoBlockBound(x, y);
        x+=blocksize;
        if (x>imgWidth-blocksize) { x=0; y+=blocksize; }
    }
}

void DrawMarkLineAt(int x, int y) {
    HDC hdc;
    x++; y++;
    hdc = GetDC(ghWnd);
    SelectObject(hdc, hpYellow);
    SelectObject(hdc, hbYellow);
    MoveToEx(hdc,   x            , y+blocksize/2-1, NULL);
    LineTo(hdc,     x+blocksize-1, y+blocksize/2-1);
    ReleaseDC(ghWnd, hdc);
    
    SelectObject(memDC, hpYellow);
    SelectObject(memDC, hbYellow);
    MoveToEx(memDC, x            , y+blocksize/2-1, NULL);
    LineTo(memDC,   x+blocksize-1, y+blocksize/2-1);
}
            
void DrawMarkAt(int x, int y, BOOL anfang) {
    HDC hdc;
    x++; y++;
    hdc = GetDC(ghWnd);
    SelectObject(hdc, hpYellow);
    SelectObject(hdc, hbYellow);
    //senkrecht mittig
    MoveToEx(hdc, x+blocksize/2-1, y+1          , NULL);
    LineTo(hdc,   x+blocksize/2-1, y+blocksize-2);
    if (anfang) {
        //waagerecht v mitte bis rechts |-
        MoveToEx(hdc, x              , y+blocksize/2-1, NULL);
        LineTo(hdc,   x+blocksize/2-1, y+blocksize/2-1);
        } else {
        //waagerecht v links bis mitte -|
        MoveToEx(hdc, x+blocksize/2-1, y+blocksize/2-1, NULL);
        LineTo(hdc,   x+blocksize  -1, y+blocksize/2-1);
    }
    ReleaseDC(ghWnd, hdc);
    
    SelectObject(memDC, hpYellow);
    SelectObject(memDC, hbYellow);
    //senkrecht mittig
    MoveToEx(memDC, x+blocksize/2-1, y+1          , NULL);
    LineTo(memDC,   x+blocksize/2-1, y+blocksize-2);
    if (anfang) {
        //waagerecht v mitte bis rechts |-
        MoveToEx(memDC, x              , y+blocksize/2-1, NULL);
        LineTo(memDC,   x+blocksize/2-1, y+blocksize/2-1);
        } else {
        //waagerecht v links bis mitte -|
        MoveToEx(memDC, x+blocksize/2-1, y+blocksize/2-1, NULL);
        LineTo(memDC,   x+blocksize  -1, y+blocksize/2-1);
    }
}

void DrawMarks(ULONGLONG Cstart, ULONGLONG Cend) {
    ULONGLONG ui=0;
    int x=0, y=0;
    
    if (blocksize < 5) return;
    if (Cstart > cDiskClusters && Cend > cDiskClusters) return;
    if (Cend   > cDiskClusters) Cend   = cDiskClusters;
    if (Cstart > cDiskClusters) Cstart = cDiskClusters;
    
    while (                          y/blocksize*clustersPL <= Cstart )
    y += blocksize;
    y -= blocksize;
    
    while ( x/blocksize*clustersPB + y/blocksize*clustersPL <= Cstart )
    x += blocksize;
    x -= blocksize;
    
    DrawMarkAt(x, y, 0);         // start mark
    x+=blocksize;
    Cstart+=clustersPB;
    
    for (ui=Cstart; ui<=Cend; ui+=clustersPB) {
        if (x>imgWidth-blocksize) { x=0; y+=blocksize; }
        if (ui+clustersPB<=Cend)
            DrawMarkLineAt(x, y);
        else
            DrawMarkAt(x, y, 1); // end mark
        x+=blocksize;
    }
}

void DrawBlockAt(int x, int y, HPEN hp, HBRUSH hb) {
    HDC hdc = GetDC(ghWnd);
    x++; y++;
    SelectObject(hdc, hp);
    SelectObject(hdc, hb);
    Rectangle(hdc, x, y, x+blocksize-1, y+blocksize-1);
    ReleaseDC(ghWnd, hdc);
    
    SelectObject(memDC, hp);
    SelectObject(memDC, hb);
    Rectangle(memDC, x, y, x+blocksize-1, y+blocksize-1);
}

void DrawBlocks(ULONGLONG Cstart, ULONGLONG Cend, HPEN hp, HBRUSH hb) {
    ULONGLONG ui=0;
    int x=0, y=0;
    
    if (Cstart > cDiskClusters && Cend > cDiskClusters) return;
    if (Cend   > cDiskClusters) Cend   = cDiskClusters;
    if (Cstart > cDiskClusters) Cstart = cDiskClusters;
    
    while (                          y/blocksize*clustersPL <= Cstart )
    y += blocksize;
    y -= blocksize;
    
    while ( x/blocksize*clustersPB + y/blocksize*clustersPL <= Cstart )
    x += blocksize;
    x -= blocksize;
    
    for (ui=Cstart; ui<Cend; ui+=clustersPB) {
        if (x>imgWidth-blocksize) { x=0; y+=blocksize; }
        DrawBlockAt(x, y, hp, hb);
        x+=blocksize;
    }
}

void DrawProgress(int progress) {
    HDC hdc;
    int x;
    if (progress == -1) {
        SelectObject(memDC, hbBG);   //delete with BG-Color
        SelectObject(memDC, hpBG);
        x = imgWidth;
        } else {
        SelectObject(memDC, hbBlue);//draw in color
        SelectObject(memDC, hpBlue);
        x = progress;
    }
    Rectangle(memDC, 0, imgHeight-5, x, imgHeight);
    hdc = GetDC(ghWnd);
    BitBlt(hdc, 0, imgHeight-5, imgWidth, imgHeight, memDC, 0, imgHeight-5, SRCCOPY);
    ReleaseDC(ghWnd, hdc);
}

void CheckForMFT(ULONGLONG *cl2, ULONGLONG *nu2) {
    ULONGLONG cluster2=*cl2, numFree2=*nu2;
    ULONGLONG cluster=0, numFree=0;
    
    int i=0;
    if (gPARTINFO.PartitionType == 7) {
        for (i=0; i<3; i++) {
            if (cluster2            <= MFTZONE[i].start &&
            cluster2 + numFree2 >  MFTZONE[i].end) {
                // MFT in the middle of free space
                cluster  = cluster2;
                numFree  = MFTZONE[i].start - cluster2;
                InsDIR("", 0);
                dirlist->Lcn = cluster;
                dirlist->Len = numFree;
                cluster2 = MFTZONE[i].end+1;
                numFree2 = (cluster2 + numFree2) - MFTZONE[i].end;
            }
            else
            if ((cluster2          >  MFTZONE[i].start) &&
            (cluster2          <  MFTZONE[i].end)       ) {
                // MFT before free space
                numFree2 = numFree2 - (MFTZONE[i].end - cluster2);
                if (numFree2 > 0)
                cluster2 = MFTZONE[i].end;
                else
                cluster2 = LLINVALID;
            }
            
            if ( cluster2 != LLINVALID &&
            (cluster2+numFree2 >  MFTZONE[i].start) &&
            (cluster2+numFree2 <  MFTZONE[i].end)       ) {
                // MFT after free space
                numFree2 = MFTZONE[i].start - cluster2;
                //cluster2 = MFTZONE[i].start;
            }
        }
    }
    if (numFree2<0) numFree2=0;
    *cl2=cluster2;
    *nu2=numFree2;
}

void InsDIRFree(ULONGLONG cluster2, ULONGLONG numFree2) {
    if (bAnalyzed) return;
    
    CheckForMFT(&cluster2, &numFree2);
    if (cluster2 != LLINVALID && numFree2 > 0 && numFree2<cDiskClusters) {
        InsDIR("", 0);
        dirlist->Lcn = cluster2;
        dirlist->Len = numFree2;
    }
}

// MoveClusters
// This uses the FSCT_MOVE_FILE interface to move the clusters of a
// file specified by the user as arguments. MoveFile requires a
// file handle, an offset within the file, the number of sectors of
// the file to move, and the target cluster on the drive to move the
// clusters to.
void MoveClusters(PMOVE_FILE_DATA pmoveFile) {
    //void MoveClusters(PMOVEFILE_DESCRIPTOR pmoveFile)
    //[file] [off] [tgt] [numclust]
    DWORD                  status;
    DWORD                  w=0;        // win32
    //IO_STATUS_BLOCK        ioStatus;
    
    // make the move
    
    // win32
    status = DeviceIoControl(ghVolume, FSCTL_MOVE_FILE,
        pmoveFile, sizeof(MOVE_FILE_DATA), NULL, 0, &w, NULL);
    
    //HandleWin32Error(GetLastError(), "FSCTL_MOVE_FILE");
    
    /*
    status = NtFsControlFile( ghVolume, NULL, NULL, 0, &ioStatus,
    FSCTL_MOVE_FILE,
    pmoveFile, sizeof( MOVEFILE_DESCRIPTOR ), NULL, 0 );
    HandleNtError( status, "FSCTL_MOVE_FILE" );
    */
    
    // If the operation is pending, wait for it to finish
    if( status == STATUS_PENDING ) {
        //_cwait(NULL, pmoveFile->FileHandle, 0);
        WaitForSingleObject( pmoveFile->FileHandle, INFINITE );
        
        // interpret the status
        if (status != STATUS_SUCCESS) {
            HandleNtError( status, "FSCTL_MOVE_FILE" );
        }
    }
    
    //if (status) pmoveFile->NumVcns = 0;
    if (!status) pmoveFile->ClusterCount = 0; // win32
}

ULONGLONG MovePartOfFile(HANDLE sourceFile, int fragNo, ULONGLONG startVcn,
ULONGLONG OldLcn, ULONGLONG NewLcn, ULONGLONG fClusters) {
    ULONGLONG        Lcn1, Lcn2;
    ULONGLONG        OldLcnFirst = OldLcn, NewLcnFirst = NewLcn;
    ULONGLONG        Offset, Clusters, movedLen=0;
    char             buff[MAX_PATH];
    //MOVEFILE_DESCRIPTOR moveFile;
    MOVE_FILE_DATA        moveFile; // win32
    
    DrawBlockBounds(OldLcnFirst, OldLcnFirst+fClusters, FALSE);
    DrawBlockBounds(NewLcnFirst, NewLcnFirst+fClusters, FALSE);
    
    Lcn1 = NewLcn;
    Lcn2 = OldLcn;
    Offset = 0;
    
    while (Offset < fClusters) {
        CheckActivities();
        if ( ghDraw != NULL ) WaitForSingleObject(ghDraw, INFINITE);
        
        Clusters = fClusters - Offset;
        if (Lcn2 > Lcn1 && Lcn2 < Lcn1 + Clusters) {
            Clusters = Lcn2 - Lcn1;
        }
        // change down clusters moved in one step to be more flexible
        if (Clusters > 128) Clusters = 128;
        
        // graphic things
        if (fragNo == 0)
        if (Lcn2 > Lcn1 && Lcn2 < Lcn1 + Clusters)
        wsprintf(buff, "shifting %I64d clusters (%I64d/%I64d) from %I64d to %I64d",
        Clusters, Offset+Clusters, fClusters, OldLcn, Lcn1);
        else if (Clusters == fClusters)
        wsprintf(buff, "moving %I64d clusters from %I64d to %I64d",
        Clusters, OldLcn, Lcn1);
        else
        wsprintf(buff, "moving %I64d clusters (%I64d/%I64d) from %I64d to %I64d",
        Clusters, Offset+Clusters, fClusters, OldLcn, Lcn1);
        else
        if (Clusters == fClusters)
        wsprintf(buff, "moving fragment #%d - %I64d clusters - from %I64d to %I64d",
        fragNo, Clusters, OldLcn, Lcn1);
        else
        wsprintf(buff, "moving fragment #%d - %I64d clusters (%I64d/%I64d) - from %I64d to %I64d",
        fragNo, Clusters, Offset+Clusters, fClusters, OldLcn, Lcn1);
        
        OutLine3(buff);
        
        // first the from-clusters in yellow
        DrawBlocks(OldLcn, OldLcn+Clusters, hpLYellow, hbLYellow);
        // now the to-clusters in L2Blue
        DrawBlocks(Lcn1, Lcn1+Clusters, hpL2Blue, hbL2Blue);
        
        ///////////////////////////////////////////////////////
        /*
        moveFile.FileHandle         = sourceFile;
        moveFile.StartVcn.QuadPart  = startVcn + Offset;
        moveFile.TargetLcn.QuadPart = Lcn1;
        moveFile.NumVcns            = (ULONG) Clusters;
        */
        // win32
        moveFile.FileHandle           = sourceFile;
        moveFile.StartingVcn.QuadPart = startVcn + Offset;
        moveFile.StartingLcn.QuadPart = Lcn1;
        moveFile.ClusterCount         = (ULONG) Clusters;
        
        MoveClusters(&moveFile);
        if (moveFile.ClusterCount < Clusters) {
            movedLen = 0;
            break;
            //if (moveFile.NumVcns < Clusters) return 0;
        }
        else movedLen += Clusters;
        ///////////////////////////////////////////////////////
        
        // now clean the from-clusters with white
        DrawBlocks(OldLcn, OldLcn+Clusters, hpYellow, hbYellow);
        // now the ready to-clusters in green
        DrawBlocks(Lcn1, Lcn1+Clusters, hpLBlue, hbLBlue);
        
        Lcn1 = Lcn1 + Clusters;
        Lcn2 = Lcn2 + Clusters;
        Offset = Offset + Clusters;
        OldLcn = OldLcn + Clusters;
        NewLcn = NewLcn + Clusters;
    }
    
    DrawBlockBounds(OldLcnFirst, OldLcnFirst+fClusters, TRUE);
    DrawBlockBounds(NewLcnFirst, NewLcnFirst+fClusters, TRUE);
    return (ULONGLONG)(movedLen);
}

ULONGLONG MoveFileNormal(char *filename, ULONGLONG Lcn,
ULONGLONG NewLcn, ULONGLONG Clusters) {
    HANDLE sourceFile;
    ULONGLONG movedLen=0;
    
    // Open the file
    sourceFile = CreateFile(filename,
    FILE_READ_ATTRIBUTES,
    FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
    NULL, OPEN_EXISTING,
    FILE_FLAG_NO_BUFFERING, 0);
    
    if( sourceFile == INVALID_HANDLE_VALUE ) {
        // try again as directory
        sourceFile = CreateFile(filename,
        GENERIC_READ,
        FILE_SHARE_READ|FILE_SHARE_WRITE,
        NULL, OPEN_EXISTING,
        FILE_FLAG_BACKUP_SEMANTICS, NULL);
        if( sourceFile == INVALID_HANDLE_VALUE ) {
            //HandleWin32Error( GetLastError(), fileName );
            return movedLen;
        }
    }
    
    // ---do the work----------------------------------
    movedLen = MovePartOfFile(sourceFile, 0, 0, Lcn, NewLcn, Clusters);
    CloseHandle(sourceFile);
    
    return movedLen;
}

// walk through the (fragmented) clusters of a file and move them to
// space starting at
ULONGLONG MoveFileFragmented (char *filename, ULONGLONG *fLcn, ULONGLONG sLen) {
    DWORD                     status;
    IO_STATUS_BLOCK           ioStatus;
    PGET_RETRIEVAL_DESCRIPTOR fileMappings;
    HANDLE                    sourceFile;
    ULONGLONG                 startVcn;
    ULONGLONG                 numClusters=0;
    ULONGLONG                 lastLcn=0, lastLen=0;
    ULONGLONG                 moved=0, movedLen=0;
    ULONGLONG                 startLcn=LLINVALID, sLcn;
    ULONGLONG                 ClusterStart=0, ClusterEnd=0, Clusters=0;
    ULONG                     fragNo=0;
    BOOL                      bdir=FALSE;
    int                       i;
    
    sLcn = *fLcn;
    if (filename[strlen(filename)-1] == '\\') bdir = TRUE;
    
    // Open the file
    sourceFile = CreateFile(filename,
    FILE_READ_ATTRIBUTES,
    FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
    NULL, OPEN_EXISTING,
    FILE_FLAG_NO_BUFFERING, 0);
    
    if( sourceFile == INVALID_HANDLE_VALUE ) {
        // try again as directory
        sourceFile = CreateFile(filename,
        GENERIC_READ,
        FILE_SHARE_READ|FILE_SHARE_WRITE,
        NULL, OPEN_EXISTING,
        FILE_FLAG_BACKUP_SEMANTICS, NULL);
        if( sourceFile == INVALID_HANDLE_VALUE ) {
            //HandleWin32Error( GetLastError(), fileName );
            return 0;
        }
    }
    
    // Start dumping the mapping information. Go until we hit EOF
    startVcn = 0;
    fileMappings = (PGET_RETRIEVAL_DESCRIPTOR) FileMap;
    while( !(status = NtFsControlFile( sourceFile, NULL, NULL, 0, &ioStatus,
    FSCTL_GET_RETRIEVAL_POINTERS,
    &startVcn, sizeof( startVcn ),
    fileMappings, FILEMAPSIZE * sizeof(LARGE_INTEGER) ) ) ||
    status == STATUS_BUFFER_OVERFLOW ||
    status == STATUS_PENDING ) {
        
        // If the operation is pending, wait for it to finish
        if( status == STATUS_PENDING ) {
            WaitForSingleObject( sourceFile, INFINITE );
            // Get the status from the status block
            if( ioStatus.Status != STATUS_SUCCESS &&
            ioStatus.Status != STATUS_BUFFER_OVERFLOW ) {
                //HandleWin32Error( GetLastError(), fileName );
                //printf("Enumerate file clusters: ");
                //PrintNtError( ioStatus.Status );
                return 0;
            }
        }
        
        // Loop through the buffer of number/cluster pairs
        startVcn = fileMappings->StartVcn;
        for( i = 0; i < (ULONGLONG) fileMappings->NumberOfPairs; i++ ) {
            // On NT 4.0, a compressed virtual run (0-filled) is
            // identified with a cluster offset of -1
            if( fileMappings->Pair[i].Lcn == LLINVALID ) {
                //printf("   VCN: %I64d VIRTUAL LEN: %I64d\n",
                //         startVcn,
                //         fileMappings->Pair[i].Vcn - startVcn );
                // compressed file - this function must quit now
                break;
                } else {
                //printf("   VCN: %I64d LCN: %I64d LEN: %I64d\n",
                //         startVcn, fileMappings->Pair[i].Lcn, fileMappings->Pair[i].Vcn - startVcn );
                
                // remember first cluster
                if (startLcn == LLINVALID) startLcn = fileMappings->Pair[i].Lcn;
                numClusters = fileMappings->Pair[i].Vcn - startVcn;
                fragNo++;
                
                // skip first frag - handle in second pass
                if (fileMappings->Pair[i].Lcn != startLcn) {
                    // handle the first frag
                    if (lastLcn == startLcn) {
                        // first cluster of a dir on FAT is not moveable
                        // so skip if next cluster is not free
                        if (bdir && gPARTINFO.PartitionType != 7) {
                            ClusterStart = sLcn;
                            Clusters = GetFreeClusters(ClusterStart+lastLen, sLen-lastLen, &ClusterStart, &ClusterEnd, TRUE);
                            if (sLcn + lastLen != ClusterStart+lastLen) {
                                break;
                            }
                        } else {
                            ClusterStart = lastLcn;
                            Clusters = GetFreeClusters(ClusterStart+lastLen, sLen-lastLen, &ClusterStart, &ClusterEnd, TRUE);
                        }
                        
                        if (sLcn + lastLen == ClusterStart &&
                            Clusters >= sLen-lastLen) 
                        {
                            // skip first fragment if enough space after it
                            ;
                        } else {
                            // else find new free space
                            ClusterStart = 0;
                            Clusters = GetFreeClusters(ClusterStart, sLen, &ClusterStart, &ClusterEnd, TRUE);
                            moved = MovePartOfFile(sourceFile, fragNo-1, 0, lastLcn, ClusterStart, lastLen);
                            
                            if( moved == lastLen ) {
                                movedLen += moved;
                                *fLcn = ClusterStart;
                                ClusterStart += moved;
                            } else {
                                CloseHandle( sourceFile );
                                return movedLen;
                            }
                        }
                    }
                    
                    // handle all following frags
                    moved = MovePartOfFile(sourceFile, fragNo,
                    startVcn, fileMappings->Pair[i].Lcn,
                    ClusterStart, numClusters);
                    
                    if( moved == numClusters ) {
                        movedLen += numClusters;
                        ClusterStart += moved;
                        } else {
                        CloseHandle( sourceFile );
                        return movedLen;
                    }
                }
                
                // now remember actual cluster
                lastLcn = fileMappings->Pair[i].Lcn;
                lastLen = numClusters;   // on NTFS?
                //lastLen = numClusters-1; // on FAT?
            }
            // move to next fragment of file
            startVcn = fileMappings->Pair[i].Vcn;
        }
        // for
        
        if ( bDriveChanged ) break;
        //fprintf(datei1, "%d\t%s\n", fragments, fileName);
        // If the buffer wasn't overflowed, then we're done
        if( !status ) break;
    }
    // while
    
    CloseHandle( sourceFile );
    return movedLen;
}

// Draw a map of all clusters in green/lightgreen/white
void DumpBitmap(void) {
    PBITMAP_DESCRIPTOR     bitMappings;
    IO_STATUS_BLOCK        ioStatus;
    DWORD                  status;
    ULONGLONG              startLcn;
    ULONGLONG              nextLcn;
    ULONGLONG              lastLcn;
    ULONGLONG              cluster;
    ULONGLONG              cluster2=LLINVALID;
    ULONGLONG              numFree=0;
    ULONGLONG              numFree2=0, numFree3=0;
    ULONGLONG              i;
    ULONGLONG              clusterC = 0;
    int                    x=0, y=0;
    
    if (!bAnalyzed) {
        // empty clear delete the dirlist
        dirlist = topdirlist;
        if (dirlist != NULL)
        DelDIR(dirlist);
        listlen = 0;
        dirlist = NULL;
        topdirlist = NULL;
        if (gPARTINFO.PartitionType == 7) {
            MFTZONE[0].start = gNTFSDATA.MftStartLcn.QuadPart;
            MFTZONE[0].end   = gNTFSDATA.MftStartLcn.QuadPart +
                               (gNTFSDATA.MftValidDataLength.QuadPart /
                                gNTFSDATA.BytesPerCluster               ); // + 1;
            MFTZONE[1].start = gNTFSDATA.MftZoneStart.QuadPart;
            MFTZONE[1].end   = gNTFSDATA.MftZoneEnd.QuadPart;
            MFTZONE[2].start = gNTFSDATA.Mft2StartLcn.QuadPart;
            MFTZONE[2].end   = gNTFSDATA.Mft2StartLcn.QuadPart + 1;
            //gNTFSDATA.MftValidDataLength.QuadPart /
            //gNTFSDATA.BytesPerCluster;
            
            // insert the first entries into dirlist
            InsDIR("$MFT", gNTFSDATA.MftValidDataLength.QuadPart);
            dirlist->fragments = 1;
            dirlist->MFT = TRUE;
            dirlist->Lcn = MFTZONE[0].start;
            dirlist->Len = MFTZONE[0].end - MFTZONE[0].start;
            
            InsDIR("$MFT2", (MFTZONE[1].end - MFTZONE[1].start)*gNTFSDATA.BytesPerCluster);
            dirlist->fragments = 1;
            dirlist->MFT = TRUE;
            dirlist->Lcn = MFTZONE[1].start;
            dirlist->Len = MFTZONE[1].end - MFTZONE[1].start;
            
            InsDIR("$MFTMirror", gNTFSDATA.BytesPerCluster);
            dirlist->fragments = 1;
            dirlist->MFT = TRUE;
            dirlist->Lcn = MFTZONE[2].start;
            dirlist->Len = MFTZONE[2].end - MFTZONE[2].start;
            } else {
            MFTZONE[0].start = 0; MFTZONE[0].end   = 0;
            MFTZONE[1].start = 0; MFTZONE[1].end   = 0;
            MFTZONE[2].start = 0; MFTZONE[2].end   = 0;
        }
    }
    // Start scanning at the cluster offset the user specifies
    bitMappings = (PBITMAP_DESCRIPTOR) BitMap;
    
    // set first cluster 'n stuff
    cluster  = 0;
    nextLcn  = 0;
    startLcn = 0;
    lastLcn  = LLINVALID;
    
    while( !(status = NtFsControlFile(
    ghVolume, NULL, NULL, 0, &ioStatus,
    FSCTL_GET_VOLUME_BITMAP, &nextLcn,
    sizeof( cluster ), bitMappings,  BITMAPSIZE)
    /*DeviceIoControl(
    ghVolume,
    FSCTL_GET_VOLUME_BITMAP,
    &nextLcn, sizeof( cluster ),
    bitMappings, BITMAPSIZE,
    zxz, &ovlp)   */
    ) ||
    status == STATUS_BUFFER_OVERFLOW ||
    status == STATUS_PENDING ) {
        // If the operation is pending, wait for it to finish
        if( status == STATUS_PENDING ) {
            WaitForSingleObject( ghVolume, INFINITE );
            // Get the status from the status block
            if( ioStatus.Status != STATUS_SUCCESS &&
            ioStatus.Status != STATUS_BUFFER_OVERFLOW ) {
                //printf("\nGet Volume Bitmap: ");
                HandleNtError( ioStatus.Status, "BitMap" );
                return;
            }
        }
        
        // Scan through the returned bitmap info, looking for empty clusters
        // starting at offset bitMappings->StartLcn
        startLcn = bitMappings->StartLcn;
        numFree = 0;
        cluster = LLINVALID;
        
        for( i = 0; i < min( bitMappings->ClustersToEndOfVol, 8*BITMAPBYTES); i++ ) {
            if( !(bitMappings->Map[ i/8 ] & BitShift[ i % 8 ])) {
                // Cluster is free -----------------
                if( cluster == LLINVALID ) {
                    cluster = startLcn + i;
                    if (cluster2 == LLINVALID)
                    cluster2 = cluster;
                    numFree = 1;
                    numFree3++;
                    } else {
                    numFree++;
                    numFree3++;
                }
                } else {
                // Cluster is not free -----------------
                if( cluster != LLINVALID ) {
                    if( lastLcn == cluster ) {
                        lastLcn = LLINVALID;
                        } else {
                        
                        numFree2 += numFree;
                        
                        if (cluster, numFree+1 < cluster2)
                        InsDIRFree(cluster, numFree);
                        if (!bAnalyzed)
                        InsDIRFree(cluster2, numFree2);
                        
                        lastLcn  = cluster;
                        numFree  = 0;
                        numFree2 = 0;
                        cluster  = LLINVALID;
                        cluster2 = LLINVALID;
                    }
                }
            }
            
            // the interesting bit
            if (++clusterC == clustersPB) {
                DrawNoBlockBound(x, y);
                if (clustersPB==1 && numFree3<1)
                DrawBlockAt(x, y, hpGreen, hbGreen);
                else if (     numFree3  >= clustersPB)
                DrawBlockAt(x, y, hpWhite, hbWhite);
                else if (numFree3 >              0     &&
                numFree3 <  clustersPB/4   )
                DrawBlockAt(x, y, hpLGreen, hbLGreen);
                else if (numFree3 >= clustersPB/4   &&
                numFree3 <  clustersPB/4*2 )
                DrawBlockAt(x, y, hpL2Green, hbL2Green);
                else if (numFree3 >= clustersPB/4*2 &&
                numFree3 <  clustersPB/4*3 )
                DrawBlockAt(x, y, hpL2Green, hbL2Green);
                else if (numFree3 >= clustersPB/4*3 &&
                numFree3 <  clustersPB     )
                DrawBlockAt(x, y, hpL3Green, hbL3Green);
                else
                DrawBlockAt(x, y, hpGreen, hbGreen);
                
                x += blocksize;
                if (x>imgWidth-blocksize) { x=0; y+=blocksize; }
                clusterC = 0;
                numFree3 = 0;
            }
            
        }
        // for all clusters in this pair
        
        if( cluster != LLINVALID && lastLcn != cluster ) {
            if (cluster, numFree+1 < cluster2) {
                if (!bAnalyzed)
                InsDIRFree(cluster, numFree);
                numFree2 = 0;
                cluster2 = cluster;
            }
            
            numFree2 += numFree;
            numFree = 0;
            cluster = LLINVALID;
        }
        
        // break at End of volume
        if( status != STATUS_BUFFER_OVERFLOW ) {
            if(cluster2 != LLINVALID) {
                if (!bAnalyzed)
                InsDIRFree(cluster2, numFree2);
                
                // interesting bit - draw last cluster
                if (numFree3 == clusterC)
                DrawBlockAt(x, y, hpWhite, hbWhite);
                else if (numFree3 >           0     &&
                numFree3 <  clusterC/4 )
                DrawBlockAt(x, y, hpLGreen, hbLGreen);
                else if (numFree3 >= clusterC/4*2 &&
                numFree3 <  clusterC/4*3 )
                DrawBlockAt(x, y, hpL2Green, hbL2Green);
                else
                DrawBlockAt(x, y, hpGreen, hbGreen);
            }
            
            // break at End of volume
            break;
        }
        // Move to the next pair
        nextLcn = bitMappings->StartLcn + i;
        
    }
    // while something in bitmappings-list
    
    if (gPARTINFO.PartitionType == 7) {
        DrawBlocks(MFTZONE[0].start, MFTZONE[0].end, hpVio, hbVio);
        DrawBlocks(MFTZONE[1].start, MFTZONE[1].end, hpVio, hbVio);
        DrawBlocks(MFTZONE[2].start, MFTZONE[2].end, hpVio, hbVio);
    }
}

// Dump the cluster of a file and insert into dirlist
// fragmented files are specific handled
struct DICT *DumpFile(char *Path, char *filename, __int64 Bytes) {
    PGET_RETRIEVAL_DESCRIPTOR   fileMappings;
    IO_STATUS_BLOCK             ioStatus;
    HANDLE                      sourceFile;
    DWORD                       status;
    ULONGLONG                   startVcn;
    ULONGLONG                   numClusters=0;
    ULONGLONG                   sLcn=0;
    ULONGLONG                   lastLcn=0;
    ULONGLONG                   lastLen=0;
    ULONGLONG                   frags=0;
    ULONGLONG                   sLen=0;
    int                         i;
    BOOL                        compressed = FALSE;
    struct DICT                 *firstFrag=NULL;
    char                        buff[MAX_PATH];
    
    // Open the file
    sourceFile = CreateFile(filename,
    FILE_READ_ATTRIBUTES,
    FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
    NULL, OPEN_EXISTING,
    FILE_FLAG_NO_BUFFERING, NULL);
    
    if( sourceFile == INVALID_HANDLE_VALUE ) {
        sourceFile = CreateFile(filename,
        GENERIC_READ,
        FILE_SHARE_READ,
        NULL, OPEN_EXISTING,
        FILE_FLAG_BACKUP_SEMANTICS, NULL);
        
        if( sourceFile == INVALID_HANDLE_VALUE ) {
            InsDIR(filename, Bytes);
            dirlist->fragments = LLINVALID;
            cLocked++;
            wsprintf(buff, "%s files (%d locked)",
            fmtNumber(cFiles, buff), cLocked);
            UpdateStatusBar(buff, 2, 0);
            return dirlist;
        }
    }
    
    // dump mapping information of file until we hit the EOF
    startVcn = 0;
    fileMappings = (PGET_RETRIEVAL_DESCRIPTOR) FileMap;
    while( !(status = NtFsControlFile( sourceFile, NULL, NULL, 0, &ioStatus,
    FSCTL_GET_RETRIEVAL_POINTERS,
    &startVcn, sizeof( startVcn ),
    fileMappings, FILEMAPSIZE * sizeof(LARGE_INTEGER) ) ) ||
    status == STATUS_BUFFER_OVERFLOW ||
    status == STATUS_PENDING ) {
        
        // If the operation is pending, wait for it to finish
        if( status == STATUS_PENDING ) {
            WaitForSingleObject( sourceFile, INFINITE );
            // Get the status from the status block
            if( ioStatus.Status != STATUS_SUCCESS &&
            ioStatus.Status != STATUS_BUFFER_OVERFLOW ) {
                //PrintNtError( ioStatus.Status );
                // some error, so make this a locked file
                InsDIR(filename, Bytes);
                dirlist->fragments = LLINVALID;
                cLocked++;
                wsprintf(buff, "%s files (%d locked)",
                fmtNumber(cFiles, buff), cLocked);
                UpdateStatusBar(buff, 2, 0);
                return dirlist;
            }
        }
        
        // Loop through the buffer of number/cluster pairs
        startVcn = fileMappings->StartVcn;
        for( i = 0; i < (ULONGLONG) fileMappings->NumberOfPairs; i++ ) {
            // On NT 4.0, a compressed virtual run (0-filled) is
            // identified with a cluster offset of -1
            if( fileMappings->Pair[i].Lcn == LLINVALID ) {
                //printf("   VCN: %I64d VIRTUAL LEN: %I64d\n",
                //  startVcn, fileMappings->Pair[i].Vcn - startVcn );
                compressed = TRUE;
                } else {
                //printf("   VCN: %I64d LCN: %I64d LEN: %I64d\n",
                //  startVcn, fileMappings->Pair[i].Lcn, fileMappings->Pair[i].Vcn - startVcn );
                
                // number of clusters from file in this pair[i]
                numClusters = fileMappings->Pair[i].Vcn - startVcn;
                sLen += numClusters;
                
                // test this cluster if it follows the lastLcn
                // this also happens on the first one because lastLcn is 0
                if ( lastLcn + lastLen != fileMappings->Pair[i].Lcn &&
                !compressed ) // yes, this file/dir is fragmented !!!
                frags++;
                
                if ( frags > 1 &&
                    fileMappings->Pair[i].Lcn != LLINVALID) {
                //&& Bytes != IINVALID ) {
                    if (frags == 2) // remember the first frag
                    firstFrag = InsDIR(filename, Bytes);
                    else            //insert frags that are greater than 2 with fragments = 0
                    InsDIR(filename, Bytes);
                    dirlist->fragmented = TRUE;
                    dirlist->fragments = 0; // will be set later for firstFrag
                    if (lastLen != 0) {
                        dirlist->Lcn = lastLcn;
                        dirlist->Len = lastLen;
                        } else {
                        dirlist->Lcn = fileMappings->Pair[i].Lcn;
                        dirlist->Len = numClusters;
                    }
                }
                
                // show fragmented clusters now in red,
                // because we lose positions
                if (fileMappings->Pair[i].Lcn != LLINVALID &&
                    frags > 1 ) {
                    if (lastLen != 0)
                    DrawBlocks(lastLcn, lastLcn+lastLen, hpRed, hbRed);
                    //DrawBlocks(fileMappings->Pair[i].Lcn,
                    //           fileMappings->Pair[i].Lcn+numClusters, hpRed, hbRed);
                }
                
                // now remember actual cluster for fragment-recognizing
                lastLcn = fileMappings->Pair[i].Lcn;
                lastLen = numClusters;   // on NTFS?
                //lastLen = numClusters-1; // on FAT?
                if (sLcn == 0)
                sLcn = lastLcn;
                
            }
            
            // goto next fragment of file
            startVcn = fileMappings->Pair[i].Vcn;
        }
        // for
        
        //fprintf(datei1, "%d\t%s\n", frags, fileName);
        // If the buffer wasn't overflowed, then we're done
        if( !status ) break;
        
    } // while something found
    CloseHandle( sourceFile );
    
    InsDIR(filename, Bytes);
    dirlist->Lcn = sLcn;
    dirlist->fragments = frags;
    dirlist->Len   = sLen;
    
    if (compressed) {
        dirlist->compressed = TRUE;
        cCompressed++;
        } else {
        if (frags > 1 && sLcn != 0) {
            dirlist->Lcn = lastLcn;
            dirlist->fragmented = TRUE;
            dirlist->fragments = 0;
            dirlist->Len = lastLen;
            
            cFragmented++;
            cFragmentsAllTogether+=(unsigned long)frags;
            
            InsDIR(filename, Bytes);
            dirlist->fragmented = TRUE;
            dirlist->fragments = frags;
            dirlist->Lcn = firstFrag->Lcn;
            dirlist->Len = sLen;
            
            wsprintf(buff, "%d fragmented", cFragmented);
            UpdateStatusBar(buff, 3, 0);
            // return the first frag, dirlist is untouched
        }
    }
    
    //(stricmp(FileName,"System Volume Information") == 0) ||
    //(stricmp(fileName,"C:/Documents and Settings/Administrator/ntuser.dat.LOG") == 0) ||
    // make these files locked
    /*
    if ( lstrcmpi(Path, cSysDir2) == 0 ) {
        int i=strlen(filename)-1, j=0;
        for (i=i; filename[i] != '\\'; i--)
        buff[j++] = filename[i];
        buff[j] = '\0'; strrev(buff);
        if( (lstrcmpi(buff, "AppEvent.Evt") == 0) ||
        (lstrcmpi(buff, "default") == 0) ||
        (lstrcmpi(buff, "default.LOG") == 0) ||
        (lstrcmpi(buff, "SAM") == 0) ||
        (lstrcmpi(buff, "SAM.LOG") == 0) ||
        (lstrcmpi(buff, "SecEvent.Evt") == 0) ||
        (lstrcmpi(buff, "SECURITY") == 0) ||
        (lstrcmpi(buff, "SECURITY.LOG") == 0) ||
        (lstrcmpi(buff, "software") == 0) ||
        (lstrcmpi(buff, "software.LOG") == 0) ||
        (lstrcmpi(buff, "SysEvent.Evt") == 0) ||
        (lstrcmpi(buff, "system") == 0) ||
        (lstrcmpi(buff, "SYSTEM.ALT") == 0) ||
        (lstrcmpi(buff, "system.LOG") == 0) ||
        (lstrcmpi(buff, "pagefile.sys") == 0) ) {
            if (firstFrag != NULL)
            firstFrag->fragments = LLINVALID;
            else
            dirlist->fragments = LLINVALID;
            cLocked++;
            wsprintf(buff, "%s files (%d locked)",
            fmtNumber(cFiles, buff), cLocked);
            UpdateStatusBar(buff, 2, 0);
        }
    }
    */

    if ( compressed )                          // compressed in ocker
        DrawBlocks(sLcn, sLcn+sLen, hpOck, hbOck);
    else if (Bytes == IINVALID)                   // dirs in blue
        DrawBlocks(sLcn, sLcn+sLen, hpBlue, hbBlue);
    else if (Bytes>sizeLimit && frags<2)
        DrawBlocks(sLcn, sLcn+sLen, hpDGreen, hbDGreen);
    else if (bShowMarks && sLen >= clustersPB*1 &&
             dirlist->fragments != LLINVALID) // large files
        DrawMarks(sLcn, sLcn+sLen);
    
    if (firstFrag != NULL)
        return firstFrag;
    
    return dirlist;
}

ULONGLONG GetFreeClusters(ULONGLONG MinLcn, ULONGLONG MinLen,
ULONGLONG *beginLcn, ULONGLONG *endLcn, BOOL MatchSpace) {
    PBITMAP_DESCRIPTOR     bitMappings;
    IO_STATUS_BLOCK        ioStatus;
    DWORD                  status;
    ULONGLONG              startLcn;
    ULONGLONG              nextLcn;
    ULONGLONG              lastLcn=LLINVALID;
    ULONGLONG              cluster;
    ULONGLONG              cluster2=LLINVALID;
    ULONGLONG              numFree;
    ULONGLONG              numFree2=0;
    ULONGLONG              i;
    
    // Start scanning at the specifies cluster offset
    bitMappings = (PBITMAP_DESCRIPTOR) BitMap;
    cluster  = 0;
    startLcn = 0;
    nextLcn  = MinLcn;
    if (nextLcn > cDiskClusters-1) return 0;
    
    while( !(status = NtFsControlFile(
    ghVolume, NULL, NULL, 0, &ioStatus,
    FSCTL_GET_VOLUME_BITMAP,
    &nextLcn, sizeof( cluster ),
    bitMappings, BITMAPSIZE)
    ) || status == STATUS_BUFFER_OVERFLOW ||
    status == STATUS_PENDING ) {
        // If the operation is pending, wait for it to finish
        if( status == STATUS_PENDING ) {
            WaitForSingleObject( ghVolume, INFINITE );
            // Get the status from the status block
            if( ioStatus.Status != STATUS_SUCCESS &&
            ioStatus.Status != STATUS_BUFFER_OVERFLOW ) {
                //printf("\nGet Volume Bitmap: ");
                HandleNtError( ioStatus.Status, "GetFreeClusters");
                return 0;
            }
        }
        
        // Scan through the returned bitmap info, looking for empty clusters
        startLcn = bitMappings->StartLcn;
        numFree = 0;
        cluster = LLINVALID;
        
        for( i = 0; i < min( bitMappings->ClustersToEndOfVol, 8*BITMAPBYTES); i++ ) {
            if( !(bitMappings->Map[ i/8 ] & BitShift[ i % 8 ])) {
                // Cluster is free -----------------
                if( cluster == LLINVALID ) {
                    cluster = startLcn + i;
                    if (cluster2 == LLINVALID)
                    cluster2 = cluster;
                    numFree = 1;
                    } else {
                    numFree++;
                }
                } else {
                // Cluster is not free ------------------
                if( cluster != LLINVALID ) {
                    
                    if( lastLcn == cluster ) {
                        lastLcn = LLINVALID;
                        } else {
                        
                        if (cluster2 + numFree2+1 < cluster) {
                            numFree2 = 0;
                            cluster2 = cluster;
                        }
                        
                        // check any left free clusters from last pair
                        if (cluster2 + numFree2+1 < cluster) {
                            CheckForMFT(&cluster2, &numFree2);
                            
                            if (cluster2 >= MinLcn && cluster2 != LLINVALID)
                            if (MatchSpace) {
                                if (numFree2 >= MinLen && cluster2 >= MinLcn) {
                                    *beginLcn = cluster2;
                                    *endLcn = cluster2 + numFree2-1;
                                    if (*endLcn == LLINVALID) *endLcn = cDiskClusters;
                                    return numFree2;
                                    } else {
                                    ; // continue
                                }
                                } else {
                                *beginLcn = cluster2;
                                *endLcn = cluster2 + numFree2-1;
                                if (*endLcn == LLINVALID) *endLcn = cDiskClusters;
                                return numFree2;
                            }
                            
                            numFree2 = 0;
                            cluster2 = cluster;
                        }
                        
                        numFree2 += numFree;
                        
                        CheckForMFT(&cluster2, &numFree2);
                        if (cluster2 >= MinLcn && cluster2 != LLINVALID)
                        if (MatchSpace) {
                            if (numFree2 >= MinLen && cluster2 >= MinLcn) {
                                *beginLcn = cluster2;
                                *endLcn = cluster2 + numFree2-1;
                                if (*endLcn == LLINVALID) *endLcn = cDiskClusters;
                                return numFree2;
                                } else {
                                ; // continue
                            }
                            } else {
                            *beginLcn = cluster2;
                            *endLcn = cluster2 + numFree2-1;
                            if (*endLcn == LLINVALID) *endLcn = cDiskClusters;
                            return numFree2;
                        }
                        
                        lastLcn  = cluster;
                        numFree  = 0;
                        numFree2 = 0;
                        cluster  = LLINVALID;
                        cluster2 = LLINVALID;
                    }
                }
            }
        }
        // for
        
        // any remaining free clusters in this pair
        if( cluster != LLINVALID && lastLcn != cluster ) {
            numFree2 += numFree;
            numFree = 0;
            cluster = LLINVALID;
        }
        
        // End of volume?
        if( status != STATUS_BUFFER_OVERFLOW ) {
            if( cluster2 != LLINVALID) {
                CheckForMFT(&cluster2, &numFree2);
                if (MatchSpace) {
                    if (numFree2  >= MinLen && cluster2 >= MinLcn) {
                        *beginLcn = cluster2;
                        *endLcn = cluster2 + numFree2-1;
                        if (*endLcn == LLINVALID)
                        *endLcn = cDiskClusters;
                        return numFree2;
                        } else {
                        ; // continue
                    }
                    } else {
                    *beginLcn = cluster2;
                    *endLcn = cluster2 + numFree2-1;
                    if (*endLcn == LLINVALID) *endLcn = cDiskClusters;
                    return numFree2;
                }
                
            }
            return 0;
        }
        
        // Move to the next block
        nextLcn = bitMappings->StartLcn + i;
    }
    // while
    
    // We only get here when there's an error
    //printf("\nGet Volume Bitmap: ");
    HandleNtError( ioStatus.Status, "GetFreeClusters" );
    return 0;
}

void LockUpDriveChange(void *dummy) {
    char strDrive[4] = "?:\\";
    
    BOOL bContinue = TRUE;
    
    strDrive[0] = cActualDrive+65;
    
    // Get a handle to a file change notification object.
    ghDriveChange = FindFirstChangeNotification (
        (LPCTSTR) strDrive, TRUE,
        FILE_NOTIFY_CHANGE_FILE_NAME |
        FILE_NOTIFY_CHANGE_DIR_NAME |
        FILE_NOTIFY_CHANGE_SIZE );
    
    // Return now if FindFirstChangeNotification failed.
    if (ghDriveChange == INVALID_HANDLE_VALUE)
    return;
    
    // Sleep until a file change notification wakes this thread or
    // m_event becomes set indicating it's time for the thread to end.
    while (bContinue) {
        if ( WaitForSingleObject(ghDriveChange, INFINITE) == WAIT_OBJECT_0 )
         {
            // Respond to a change notification.
            if (ghDefragFragments != NULL || ghDefragFreeSpace != NULL)
                PostMessage (ghWnd, WM_USER, /*(WPARAM)*/ WM_DRIVECHANGED, 0);

            bContinue = FALSE;
        }
    }
    
    // Close the file change notification handle
    if (ghDefragFragments != NULL && ghDefragFreeSpace != NULL)
    FindCloseChangeNotification (ghDriveChange);
    
    return;
}

/* Scan a directory and all it's subdirectories (recursive) for files
creates a list of files and their locations in memory for later use*/
void ScanDir(char *Path) {
    HANDLE FindHandle;
    WIN32_FIND_DATA FFData;
    char SearchPath[MAX_PATH];
    char FileName[MAX_PATH];
    char buff[MAX_PATH];
    __int64 Bytes;
    
    sprintf(SearchPath, "%s*", Path);
    FindHandle = FindFirstFile(SearchPath, &FFData);
    if (FindHandle == INVALID_HANDLE_VALUE) return;
    do {
        // let App repaint
        if ( ghDraw != NULL ) WaitForSingleObject(ghDraw, INFINITE);
        
        if (strcmp(FFData.cFileName,".") == 0) continue;
        if (strcmp(FFData.cFileName,"..") == 0) continue;
        
        if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0) {
            // it is a dir
            sprintf(FileName, "%s%s\\", Path, FFData.cFileName);
            
            cDirs++;
            if (subDirLevel == 0) cRootDirs++;
            subDirLevel++;
            
            // wait for draw or break;
            if ( ghDraw != NULL ) WaitForSingleObject(ghDraw, INFINITE);
            if ( bStopAction ) break;
            if ( bDriveChanged ) break;
            
            /* Analyze the item: number of clusters, LCN (the physical location on
            harddisk), and number of fragments. */
            DumpFile(Path, FileName, IINVALID);
            
            if (cDirs % 16 == 0) {
                // Update the status bar
                wsprintf(buff, "%s dirs", fmtNumber(cDirs, buff));
                UpdateStatusBar(buff, 1, 0);
                if (cLocked>0)
                    wsprintf(buff, "%s files (%d locked)", fmtNumber(cFiles, buff), cLocked);
                else
                    wsprintf(buff, "%s files", fmtNumber(cFiles, buff));
                UpdateStatusBar(buff, 2, 0);
                wsprintf(buff, "%s byte", fmtNumber(cBytes, buff));
                UpdateStatusBar(buff, 4, 0);
            }
            
            if (subDirLevel<5) {
                wsprintf(buff, "Analyzing %s", FileName);
                OutLine3(buff);
                //UpdateStatusBar(buff, 4, 0);
            }
            
            // recurse Call to ScanDir
            ScanDir( FileName ); //
            
            } else {
            // it is a file
            sprintf(FileName,"%s%s", Path, FFData.cFileName);
            Bytes = FFData.nFileSizeHigh *
            ((LONGLONG)MAXDWORD + 1) + FFData.nFileSizeLow;
            
            cFiles++;
            cBytes+=Bytes;
            
            // Show progress
            progress = (long int)( ((float)(cBytes)/(float)(cBytesUsed))*imgWidth );
            if (progress > lastprogress) {
                Sleep(0);
                lastprogress = progress;
                DrawProgress(progress);
            }
            
            // wait for draw or break
            if ( ghDraw != NULL ) WaitForSingleObject(ghDraw, INFINITE);
            if ( bStopAction ) break;
            if ( bDriveChanged ) break;
            
            /* Analyze the item: number of clusters, LCN (the physical location on
            harddisk), and number of fragments. */
            DumpFile(Path, FileName, Bytes);
            
            /* Ignore the item if it has no LCN and no clusters. Very small
            files are stored in the MFT Master File Table and are reported by Windows as
            having zero LCN and clusters. */
            //if ((Lcn == 0) && (Clusters == 0))
            //    continue;
            
            /* Ignore compressed files. For some reason or other they're always
            reported as fragmented. */
            //if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED) != 0)
            //    continue;
            
            /* Show info about the file. */
            /*
            fprintf(stdout,"Defragmenting: %s\n",FileName);
            if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED) != 0) {
                fprintf(stdout,"  Special file attribute: Compressed\n");
            }
            if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED) != 0) {
                fprintf(stdout,"  Special file attribute: Encrypted\n");
            }
            if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_OFFLINE) != 0) {
                fprintf(stdout,"  Special file attribute: Offline\n");
            }
            if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_READONLY) != 0) {
                fprintf(stdout,"  Special file attribute: Read-only\n");
            }
            if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0) {
                fprintf(stdout,"  Special file attribute: Reparse-point\n");
            }
            if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_SPARSE_FILE) != 0) {
                fprintf(stdout,"  Special file attribute: Sparse-file\n");
            }
            if ((FFData.dwFileAttributes & FILE_ATTRIBUTE_TEMPORARY) != 0) {
                fprintf(stdout,"  Special file attribute: Temporary\n");
            }
            fprintf(stdout,"  %lu clusters at %I64d, %I64d bytes, %lu fragments\n",
            Clusters,Lcn,Bytes,Fragments);
            */
        }
    }
    while (FindNextFile(FindHandle,&FFData) != 0);
    
    subDirLevel--;
    FindClose(FindHandle);
}

void DrawBitMap(void *dummy) {
    ULONGLONG lastLcn = cDiskClusters;
    unsigned long c=0;
    char laststr[MAX_PATH];
    struct DICT *Here=NULL, *merk=NULL;
    int myprogress=0, mylastprogress=0;
    
    strcpy(laststr, LastOutLine3);
    OutLine3("Drawing Disk Layout...");

    // draw diskmap
    DumpBitmap();
    
    if (topdirlist != NULL)
    // make a copy if something is running
    if (ghAnalyze || ghDefragFragments || ghDefragFreeSpace || ghDefrag)
    Here = merk = SortDict(CopyDIR(topdirlist), listlen, byLcnDesc);
    else
    Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
    
    // first the system files
    while (Here != NULL && !ghAnalyze && bAnalyzed) {
        progress = (long int)( (float)c++/(float)(listlen*3*imgWidth));
        if (progress > lastprogress) {
            lastprogress = progress;
            DrawProgress(progress);
        }
        if (Here->fragments == LLINVALID && Here->Len > 0 && !Here->fragmented)
        DrawBlocks(Here->Lcn, Here->Lcn+Here->Len, hpBrown, hbBrown);
        
        if (Here->Lcn < cDiskClusters &&
        Here->Lcn >=0 &&
        !ghDefragFragments && !ghDefragFreeSpace) {
            // check if there is a hole
            if (Here->Lcn+Here->Len+1 <= lastLcn &&
            ! Here->free && ! Here->fragmented)
            DrawBlocks(Here->Lcn+Here->Len,
            (Here->Lcn+Here->Len) +
            (lastLcn - (Here->Lcn+Here->Len)), hpBrown, hbBrown);
            
            lastLcn = Here->Lcn;    // remember last position
        }
        Here = Here->next;
    }
    
    if (merk != NULL) Here = merk;
    else Here = topdirlist;
    // now draw the other files
    while (Here != NULL && !ghAnalyze && bAnalyzed) {
        myprogress = (long int)( (float)c++/(float)(listlen*3*imgWidth));
        if (myprogress > mylastprogress) {
            mylastprogress = myprogress;
            DrawProgress(myprogress);
        }
        if (Here->filesize == IINVALID &&
        !Here->fragmented)                    // this is a dir
        DrawBlocks(Here->Lcn, Here->Lcn + Here->Len, hpBlue, hbBlue);
        else if (Here->compressed)                // compressed
        DrawBlocks(Here->Lcn, Here->Lcn + Here->Len, hpOck, hbOck);
        else if (Here->filesize>sizeLimit &&      // above sizeLimit
        Here->fragments!=LLINVALID && ! Here->MFT && ! Here->fragmented && ! Here->free)
        DrawBlocks(Here->Lcn, Here->Lcn + Here->Len, hpDGreen, hbDGreen);
        
        if ( bDriveChanged ) break;
        Here = Here->next;
    }
    
    // do again the fragmented files
    if (merk != NULL) Here = merk;
    else Here = topdirlist;
    while (Here != NULL) {
        progress = (long int)( (float)c++/(float)(listlen*3*imgWidth) );
        if (progress > lastprogress) {
            lastprogress = progress;
            DrawProgress(progress);
        }
        if (Here->fragmented           &&    // fragmented
            Here->fragments == 0)            // and not the description-entry in dirlist
        DrawBlocks(Here->Lcn, Here->Lcn + Here->Len, hpRed, hbRed);
        
        if (Here->Len > clustersPB*1   &&    // large files
            Here->fragments < 2        &&    // and not the description-entry in dirlist
            ! Here->free && bShowMarks)      // and not free space
        DrawMarks(Here->Lcn, Here->Lcn + Here->Len);
        
        Here = Here->next;
    }
    
    if (IsWindowVisible(ghListWnd))
    DrawBlockBound(xposlast, yposlast);
    
    if (merk != NULL) DelDIR(merk);
    
    OutLine3(laststr);
    
    DrawProgress(-1);
    
    ghDraw = NULL;
}

void DumpLayoutToFile(void) {
    FILE *logfile;
    struct DICT *Here;
    char buff[1024];
    unsigned long c=0;

    logfile = fopen("c:\\def2.log", "w");

    DrawProgress(-1);
    progress=0, lastprogress=0;

    OutLine3("dump to file");
    Here = topdirlist = SortDict(topdirlist, listlen, byLcnAsc);
    fprintf(logfile, "StartLcn   EndLcn     Len        fragments  FileSize     dir/file\n");
    while (Here != NULL) {
        progress = (long int)( (float)c++/(float)(listlen*imgWidth)*1000000);
        if (progress > lastprogress) {
            lastprogress = progress;
            DrawProgress(progress);
        }

        wsprintf(buff, "%10I64d %10I64d %10I64d %5I64d %12I64d %s",
            Here->Lcn, Here->Lcn+Here->Len, Here->Len,
            Here->fragments, Here->filesize, Here->fileName);
        if (Here->free) lstrcat(buff, "---free space---\n");
                   else lstrcat(buff, "\n");
        fprintf(logfile, buff);
        Here = Here->next;
    }
    OutLine3("");

    fclose (logfile);
}

void AnalyzeVolume(void *dummy) {
    char buff[MAX_PATH];
    HMENU hsub;
    ULONGLONG lastLcn = cDiskClusters;
    unsigned long c=0;
    struct DICT *Here;
    
    // maybe there has changed something
    GetDiskParameters();
    
    bDriveChanged = FALSE;
    bAnalyzed = FALSE;
    
    // draw empty diskmap
    DumpBitmap();
    if (!bStopAction && !bDriveChanged) bAnalyzed = TRUE;
    
    DrawProgress(-1);
    
    buff[0]=cActualDrive+65; buff[1]=':'; buff[2]='\\'; buff[3]='\0';
    
    cRootDirs=0; cDirs=0; cFiles=0; cBytes=0;
    cLocked=0; cCompressed=0; cFragmented=0; cFragmentsAllTogether=0;
    subDirLevel=0;
    
    progress=0, lastprogress=0;
    
    // ---------------------------------
    //DumpFile(buff, "$Bitmap", 0);
    //dirlist->MFT = TRUE;

    /*$AttrDef, $BadClus, $Bitmap, $Boot, $LogFile, 
      $MFT, $MFTMirr, $MFT::$Bitmap, $Secure, $UpCase, $Volume und $Quota*/
    
    // create dirlist - analyze dirs and files
    DumpFile(buff, buff, IINVALID); // root dir
    ScanDir(buff);
    
    //DumpLayoutToFile();
    progress=0, lastprogress=0;
    
    // draw the system files if any
    Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
    while (Here != NULL && !bStopAction && !bDriveChanged) {
        progress = (long int)( (float)c++/(float)(listlen*2*imgWidth));
        if (progress > lastprogress) {
            lastprogress = progress;
            DrawProgress(progress);
        }
        if (Here->fragments == LLINVALID && Here->Len > 0)
        DrawBlocks(Here->Lcn, Here->Lcn+Here->Len, hpBrown, hbBrown);
        
        if (Here->Lcn < cDiskClusters &&
        Here->Lcn >=0) {
            // check if there is a hole
            if (Here->Lcn+Here->Len+1 < lastLcn &&
            ! Here->free && ! Here->fragmented)
            DrawBlocks(Here->Lcn+Here->Len,
            (Here->Lcn+Here->Len) +
            (lastLcn - (Here->Lcn+Here->Len)), hpBrown, hbBrown);
            
            lastLcn = Here->Lcn;    // remember last position
        }
        Here = Here->next;
    }
    
    // do again the fragmented files
    Here = topdirlist;
    while (Here != NULL && !bStopAction && !bDriveChanged) {
        progress = (long int)( (float)c++/(float)(listlen*2*imgWidth) );
        if (progress > lastprogress) {
            lastprogress = progress;
            DrawProgress(progress);
        }
        if (//Here->filesize != IINVALID      &&    // this is a file
        Here->fragmented                &&    // fragmented
        Here->fragments == 0)                 // and not the description-entry in dirlist
        DrawBlocks(Here->Lcn, Here->Lcn + Here->Len, hpRed, hbRed);
        
        if (Here->Len > clustersPB*1   &&    // large files
        Here->fragments > 0        &&    // and not the description-entry in dirlist
        ! Here->free && bShowMarks)      // and not free space
        DrawMarks(Here->Lcn, Here->Lcn + Here->Len);
        
        Here = Here->next;
    }
    
    DrawProgress(-1);
    OutLine3("");
    
    if (ghDriveChange != NULL) {
        FindCloseChangeNotification(ghDriveChange);
        ghDriveChange = NULL;
    }
    
    if ( bStopAction ) {
        bAnalyzed = FALSE;
        // draw diskmap
        DumpBitmap();

        ResetStatusBar();
        UpdateStatusBar("Analyze stopped", 4, 0);
    } else {
        cDirsFiles = cDirs + cFiles;
        
        // Update the status bar
        wsprintf(buff, "%s dirs", fmtNumber(cDirs, buff));
        UpdateStatusBar(buff, 1, 0);
        if (cLocked>0)
            wsprintf(buff, "%s files (%d locked)", fmtNumber(cFiles, buff), cLocked);
        else
            wsprintf(buff, "%s files", fmtNumber(cFiles, buff));
        UpdateStatusBar(buff, 2, 0);
        wsprintf(buff, "%d fragmented", cFragmented);
        UpdateStatusBar(buff, 3, 0);
        wsprintf(buff, "%s byte", fmtNumber(cBytes, buff));
        UpdateStatusBar(buff, 4, 0);
    }
    
    wsprintf(buff, "&Analyze\tF5");
    hsub = GetSubMenu(GetMenu(ghWnd), 0);
    
    DeleteMenu(hsub, 0, MF_BYPOSITION);
    InsertMenu(hsub, 0, MF_STRING|MF_BYPOSITION|MF_ENABLED, IDM_ANALYZE, buff);
    if (ghDefrag == NULL && ghDefragFragments == NULL && ghDefragFreeSpace == NULL) {
        EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 2, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 3, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 5, MF_BYPOSITION|MF_ENABLED);
        hsub = GetMenu(ghWnd);
        EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
        } else {
        EnableMenuItem(hsub, 0, MF_BYPOSITION|MF_GRAYED);
    }
    DrawMenuBar(ghWnd);
    
    ghAnalyze = NULL;
    
    if (IsWindowVisible(ghListWnd)) {
        GetWindowText(ghListWnd, buff, MAX_PATH);
        if (f_instr("Drive Information", buff) >= 0)
        UpdateDriveInformation();
        else
        UpdateClusterList();
    }
}

void DefragFragments(void *dummy) {
    struct DICT *Here, *New;
    char buff[MAX_PATH], buff2[MAX_PATH];
    unsigned long c=0;
    ULONGLONG lastLcn=0;
    HMENU hsub;
    unsigned long FilesMoved=0;
    
    // wait for possible analyze
    if (!CheckActivities()) {

        UpdateStatusBar("Defragging...", 4, 0);
        
        progress=0; lastprogress=0;
        
        Here = topdirlist = SortDict(topdirlist, listlen, bySizeAsc);
        //Here = topdirlist = SortDict(topdirlist, listlen, bySizeDesc);

        while (Here != NULL && cFragmented>0 ) {
            
            progress = (long int)( (float)c++/(float)(listlen)*imgWidth );
            if (progress > lastprogress) {
                Sleep(0);
                lastprogress = progress;
                DrawProgress(progress);
            }
            
            if (Here->fragments != LLINVALID  &&  // skip locked files
                //Here->filesize != IINVALID    &&  // skip dirs
                Here->Len > 0                 &&  // skip null-files
                Here->filesize != 0           &&
                Here->filesize < sizeLimit    &&
                ! Here->free                  &&
                ! Here->compressed            &&  // skip compressed files
                ! Here->MFT                   &&
                Here->fragmented              &&  // fragmented !!!
                Here->fragments > 1) 
            {

                if (CheckActivities()) break;
                
            if (Here->filesize != IINVALID)
                   sprintf(buff, "file: %.3f MB - %I64d fragments - %s",
                      (float)Here->filesize/1024/1024, Here->fragments, Here->fileName);
                else 
                   sprintf(buff, "dir: %I64d fragments - %s", Here->fragments, Here->fileName);
                UpdateStatusBar(buff, 4, 0);

                //if (gPARTINFO.PartitionType != 7)
                
                if (MoveFileFragmented(Here->fileName, &Here->Lcn, Here->Len) == Here->Len) {
                    FilesMoved++;
                    lastLcn = Here->Lcn;
                    //InsDIRFree(Here->Lcn, Here->Len);
                    //Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
                    New = DumpFile("", Here->fileName, Here->filesize);
                    //Here = DumpFile(Here->fileName, Here->filesize);
                    if (!New->fragmented) {
                        Here->fragmented = FALSE;
                        Here->fragments = 1;
                        wsprintf(buff, "%d fragmented", --cFragmented);
                        UpdateStatusBar(buff, 3, 0);
                        OutLine3("");
                    }
                } else {
                    wsprintf(buff2, "not defragged: %s", Here->fileName);
                    OutLine3(buff2);
                    UpdateStatusBar(buff2, 4, 0);
                }
            } // if
            Here = Here->next;
        } // while Here != NULL
    } // if (CheckActivities()) break;
    
    DrawProgress(-1);
    
    ghDefragFragments = NULL;
    if (ghDriveChange != NULL) {
        FindCloseChangeNotification(ghDriveChange);
        ghDriveChange = NULL;
    }
    
    wsprintf(buff, "Defrag &Fragmented\tF7");
    hsub = GetSubMenu(GetMenu(ghWnd), 0);
    DeleteMenu(hsub, 1, MF_BYPOSITION);
    InsertMenu(hsub, 1, MF_STRING|MF_BYPOSITION|MF_ENABLED, IDM_DEFRAGFRAGS, buff);
    // re-enable menu
    if (ghDefrag == NULL) {
        EnableMenuItem(hsub, 0, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 2, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 3, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 5, MF_BYPOSITION|MF_ENABLED);
        hsub = GetMenu(ghWnd);
        EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
        } else {
        EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_GRAYED);
    }
    DrawMenuBar(ghWnd);
    
    wsprintf(buff, "");
    if (FilesMoved > 0) {
        wsprintf(buff, "%s file(s) moved", fmtNumber(FilesMoved, buff));
        ResetStatusBar();
        
        bAnalyzed = FALSE;
        // draw disk-layout
        DrawBitMap(NULL);
        // do an new Analyze
        //SendMessage(ghWnd, WM_COMMAND, IDM_ANALYZE, 0);
    }
    else if (cFragmented!=0) 
        wsprintf(buff, "dirs or files above limit fragmented.");
    else if (cFragmented==0)
        wsprintf(buff, "Nothing to defrag.");

    OutLine3(buff);
    UpdateStatusBar("DefragFragmented finished", 4, 0);
}

void DefragFreeSpace(void *dummy) {
    struct DICT *Here, *BestFit;
    ULONGLONG ClusterStart=0, ClusterEnd=0, Clusters=1;
    ULONGLONG fsCS=0, fsCE=0; // freeSpace clusterindicators
    ULONGLONG cfCS=0, cfCE=0; // currentfile clusterindicators
    ULONGLONG lokalCS=0, lokalCE=0;
    ULONGLONG cBound=0;
    char buff[MAX_PATH];
    HMENU hsub;
    int i;
    unsigned long FilesMoved=0;
    
    // wait for possible analyze
    if (!CheckActivities()) {
        
        //OutLine2("Optimizing...");
        UpdateStatusBar("Optimizing...", 4, 0);
        
        progress=0, lastprogress=0;
        // normal files starting from end of volume
        
        //Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
        Here = topdirlist = SortDict(topdirlist, listlen, byLcnAsc);

        // first run
        while (1) {
            if (CheckActivities()) break;
            
            // find the next block with free space
            Clusters = GetFreeClusters(fsCS, 1,
            &ClusterStart, &ClusterEnd, FALSE);
            if (Clusters == 0) break;
            fsCS = ClusterStart;
            cfCS = ClusterEnd;
            
            i = 0;
            while (ClusterStart <= ClusterEnd && Clusters > 0) {
                if (CheckActivities()) break;
                
                BestFit = NULL;
                // Find a file above ClusterStart
                Here = topdirlist;
                while ( Here != NULL ) {
                    if (Here->filesize >= 0           &&
                    (
                    Here->filesize < sizeLimit    || // limit filesize in bytes
                    (Here->filesize == IINVALID   &&
                    gPARTINFO.PartitionType == 7)
                    )                             &&
                    Here->Lcn >= cfCS             && // a file above current clusterbound
                    Here->Len > 0                 && // skip zero byte files
                    Here->Len <= Clusters         &&
                    //Here->fragments != LLINVALID  && // skip locked files
                    ! Here->free                  &&
                    ! Here->fragmented            &&
                    ! Here->MFT                   &&
                    ! Here->compressed )             // skip compressed files
                    {
                        if (BestFit == NULL)
                        BestFit = Here;
                        if (Here->filesize == IINVALID &&
                        gPARTINFO.PartitionType == 7) {
                            if (BestFit->filesize != IINVALID ||
                            Here->Len >= BestFit->Len)
                            BestFit = Here;
                            } else {
                            if (BestFit->filesize != IINVALID &&
                            Here->Len >= BestFit->Len)
                            BestFit = Here;
                        }
                    }
                    Here = Here->next;
                }
                Here = BestFit;
                if (Here == NULL) break;
                
                // set new currentFile clusterStart
                //if (Here->Lcn > 0 && Here->Lcn != LLINVALID)
                //    cfCS = Here->Lcn + Here->Len;
                
                // skip dirs on FAT
                if (Here->filesize == IINVALID && gPARTINFO.PartitionType != 7 ) {
                    // mark it as unmoveable
                    Here->fragments = LLINVALID;
                    continue;
                }
                
                if (CheckActivities()) break;
                
                // skip wrong files
                if (Here->Len > Clusters-1 &&
                Here->Lcn+Here->Len < ClusterStart)
                break;
                
            if (Here->filesize != IINVALID)
                   sprintf(buff, "%.3f MB - %s",
                      (float)Here->filesize/1024/1024, Here->fileName);
                else 
                   sprintf(buff, "%s", Here->fileName);
                UpdateStatusBar(buff, 4, 0);
                
                if (MoveFileNormal(Here->fileName, Here->Lcn,
                    ClusterStart, Here->Len) == Here->Len) {
                    FilesMoved++;
                    Here->Lcn = ClusterStart;
                    Clusters -= Here->Len;
                    fsCS += Here->Len;
                    //cfCS += Here->Len;
                    ClusterStart = Here->Lcn+Here->Len;
                    //InsDIRFree(Here->Lcn, Here->Len);
                    //Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
                }
                else break;
                
                //cBound = ClusterStart + Here->Len;
                //cfCS = cfCS + Here->Len;
                i++;
            }
            // while
            if (i > 0) continue;
            
            i = 0;
            while (1) {
                // Find a File above ClusterEnd
                Here = topdirlist;
                while ( Here != NULL ) {
                    if (//Here->Len <= Clusters         &&
                    Here->filesize < sizeLimit    && // limit filesize in bytes
                    Here->Lcn == cfCS           && // a file behind the free cluster
                    Here->Len > 0                 && // skip zero byte files
                    //Here->fragments != LLINVALID  && // skip locked files
                    //--------------------------------------------------
                    //Here->filesize == IINVALID    && // skip dirs
                    ! Here->free                  &&
                    ! Here->fragmented            &&
                    ! Here->MFT                   &&
                    ! Here->compressed )             // skip compressed files
                    break;
                    Here = Here->next;
                }
                if (Here == NULL) {
                    fsCS = cfCS;
                    //ClusterStart = ClusterEnd+1;
                    break;
                }
                
                if (Here->Len > 100) {
                    // move the file away
                    if (CheckActivities()) break;
                    
                    if (GetFreeClusters(0, Here->Len,
                        &lokalCS, &lokalCE, TRUE) >= Here->Len)
                    {
                        if (CheckActivities()) break;
                        
                        if (Here->filesize == IINVALID && gPARTINFO.PartitionType != 7 )
                           continue;

                  if (Here->filesize != IINVALID)
                         sprintf(buff, "%.3f MB - %s",
                            (float)Here->filesize/1024/1024, Here->fileName);
                      else 
                         sprintf(buff, "%s", Here->fileName);
                        UpdateStatusBar(buff, 4, 0);
                        
                        if (MoveFileNormal(Here->fileName, Here->Lcn,
                            lokalCS, Here->Len) == Here->Len)
                        {
                            FilesMoved++; i++;
                            Here->Lcn = lokalCS;
                            if (lokalCS == ClusterStart)
                            ClusterStart = lokalCS + Here->Len;
                            //Clusters -= Here->Len;
                            //cfCS = Here->Lcn+Here->Len;
                            //InsDIRFree(Here->Lcn, Here->Len);
                            //Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
                            //lokalCS = Here->Lcn + Here->Len;
                            //ClusterEnd = ClusterEnd + Here->Len;
                            break;
                        }
                        else break;
                    }
                }
                
                if (CheckActivities()) break;
                
                if (Here->filesize == IINVALID && gPARTINFO.PartitionType != 7 )
                break;
                
            if (Here->filesize != IINVALID)
                    sprintf(buff, "%.3f MB - %s",
                       (float)Here->filesize/1024/1024, Here->fileName);
            else 
                     sprintf(buff, "%s", Here->fileName);
                UpdateStatusBar(buff, 4, 0);
                
                if (MoveFileNormal(Here->fileName, Here->Lcn,
                    ClusterStart, Here->Len) == Here->Len)
                {
                    FilesMoved++;
                    Here->Lcn = ClusterStart;
                    Clusters -= Here->Len;
                    fsCS = ClusterStart + Here->Len;
                    ClusterStart += Here->Lcn+Here->Len;
                    //InsDIRFree(Here->Lcn, Here->Len);
                    //Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
                }
                else break;
                
////////////////////////////////////
                //cfCS = cfCS + Here->Len;
                //ClusterStart = ClusterStart + Here->Len;
                //ClusterEnd = ClusterEnd + Here->Len;
                
                i++;
            }
            // while find a file above
            
            if (i > 0) continue;
            fsCS = fsCS + 1;
            //ClusterStart = ClusterEnd;
            
        }
        // while 1
        
    }
    // if (CheckActivities()) break;
    
    ghDefragFreeSpace = NULL;
    if (ghDriveChange != NULL) {
        FindCloseChangeNotification(ghDriveChange);
        ghDriveChange = NULL;
    }
    
    wsprintf(buff, "Defrag Free &Space\tF8");
    hsub = GetSubMenu(GetMenu(ghWnd), 0);
    DeleteMenu(hsub, 2, MF_BYPOSITION);
    InsertMenu(hsub, 2, MF_STRING|MF_BYPOSITION|MF_ENABLED, IDM_DEFRAGFREE, buff);
    // re-enable menu
    if (ghDefrag == NULL) {
        EnableMenuItem(hsub, 0, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 3, MF_BYPOSITION|MF_ENABLED);
        EnableMenuItem(hsub, 5, MF_BYPOSITION|MF_ENABLED);
        hsub = GetMenu(ghWnd);
        EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
        } else {
        EnableMenuItem(hsub, 2, MF_BYPOSITION|MF_GRAYED);
    }
    DrawMenuBar(ghWnd);
    
    wsprintf(buff, "");
    if (FilesMoved > 0) {
        wsprintf(buff, "%s file(s) moved", fmtNumber(FilesMoved, buff));
        ResetStatusBar();
        
        bAnalyzed = FALSE;
        // draw disk-layout
        DrawBitMap(NULL);
        // do an new Analyze
        //SendMessage(ghWnd, WM_COMMAND, IDM_ANALYZE, 0);
    }
    //OutLine2(buff);
    OutLine3(buff);
    UpdateStatusBar("DefragFreeSpace finished", 4, 0);
}

void Defrag (void *dummy) {
    char buff[MAX_PATH];
    HMENU hsub;
    
    SendMessage(ghWnd, WM_COMMAND, IDM_DEFRAGFRAGS, 0);
    if ( ghDefragFragments != NULL )
    WaitForSingleObject(ghDefragFragments, INFINITE);
    
    // wait for possible analyze
    if (!CheckActivities()) {
        SendMessage(ghWnd, WM_COMMAND, IDM_DEFRAGFREE, 0);
        if ( ghDefragFreeSpace != NULL )
        WaitForSingleObject(ghDefragFreeSpace, INFINITE);
    }
    
    ghDefrag = NULL;
    
    wsprintf(buff, "&Defrag Fragmented && Free Space\tF9");
    hsub = GetSubMenu(GetMenu(ghWnd), 0);
    DeleteMenu(hsub, 3, MF_BYPOSITION);
    InsertMenu(hsub, 3, MF_STRING|MF_BYPOSITION|MF_ENABLED, IDM_DEFRAG, buff);
    EnableMenuItem(hsub, 0, MF_BYPOSITION|MF_ENABLED);
    EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
    EnableMenuItem(hsub, 2, MF_BYPOSITION|MF_ENABLED);
    EnableMenuItem(hsub, 5, MF_BYPOSITION|MF_ENABLED);
    hsub = GetMenu(ghWnd);
    EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
    DrawMenuBar(ghWnd);

    UpdateStatusBar("Defrag finished", 4, 0);
}

// this function is why I made this
void DefragByDirectories(void *dummy) {
    struct DICT *Here, *AList, *ALtop, *ALthisdir=NULL;
    ULONGLONG ClusterStart=0, ClusterEnd=0, Clusters=0;
    ULONGLONG al_size, endDirLcn=0, lastDirLcn=0, lastDiskLcn=0;
    ULONGLONG dirsize=0, dirsizecounter=0;
    ULONGLONG cc=0;
    HMENU hsub;
    int i, lenDir;
    char buff[MAX_PATH];
    char DirToMove[MAX_PATH], DirToMoveLast[MAX_PATH], DirToMoveFirst[MAX_PATH];
    unsigned long FilesMoved=0, RootDirsCounter=0;
    
    // wait for possible analyze
    if (!CheckActivities()) {
        
        UpdateStatusBar("Optimizing...", 4, 0);
        progress=0, lastprogress=0;
        
        // debug - set only this dir as working
        if (cActualDrive == cWinDir[0]-65) {
            lstrcpy((LPTSTR)DirToMoveFirst, (LPCTSTR)cWinDir);
            } else {
            //InputBox MessageBox
            lstrcpy((LPTSTR)DirToMoveFirst, (LPCTSTR)"C:\\WIN98\\");
        }
        
        lstrcpy((LPTSTR)DirToMove, (LPCTSTR)DirToMoveFirst);
        lstrcpy((LPTSTR)DirToMoveLast, (LPCTSTR)DirToMoveFirst);
        
        // copy to a new dirlist for better handling
        AList = ALtop = SortDict(CopyDIR(topdirlist), listlen, byName);
        Here = topdirlist = SortDict(topdirlist, listlen, byLcnAsc);
        while (AList != NULL) {
            if (CheckActivities()) break;
            
            if (RootDirsCounter==cRootDirs) break;
            
            // check if this dir is finished
            if ( _strnicmp(DirToMoveFirst, DirToMove, strlen(DirToMove)) != 0) {
                while (AList != NULL ) {
                    if (//AList->filesize == IINVALID &&
                    _strnicmp(DirToMoveLast, AList->fileName,
                    strlen(DirToMoveLast) ) != 0 ) {
                        // dir is finished - build next path
                        for (i=0; ; i++) {
                            DirToMove[i] = AList->fileName[i];
                            if ((DirToMove[i]=='\\' || DirToMove[i]=='\0')
                            && i>3) {
                                DirToMove[i+1] = '\0';
                                break;
                            }
                            ;
                            //strrchr(string, ch) - string + 1;
                        }
                        
                        if (_strnicmp(DirToMove, DirToMoveLast, lenDir) != 0) {
                            dirsizecounter = 0;
                            break; // found next dir
                        }
                    }
                    AList = AList->next;
                }
            }
            ALthisdir = AList;
            
            RootDirsCounter++;
            lstrcpy(DirToMoveLast, DirToMove);
            lenDir = strlen(DirToMove);
            
            // count size of DirToMove
            dirsize = 0;
            AList = ALtop;
            while (AList != NULL ) {
                if (//AList->filesize != IINVALID   && // skip dirs
                AList->Len > 0                && // skip zero byte files
                AList->fragments != LLINVALID && // skip locked
                ! AList->free                 &&
                ! AList->fragmented           && // skip fragmented
                ! AList->compressed           && // skip compressed
                ! AList->MFT                  &&
                _strnicmp(AList->fileName, DirToMove, lenDir) == 0 ) {
                    dirsize += AList->Len;
                }
                AList = AList->next;
            }
            
            Clusters = 0; ClusterStart = ClusterEnd = 0;
            Clusters = GetFreeClusters(ClusterStart, 1,
            &ClusterStart, &ClusterEnd, TRUE);
            lastDirLcn = ClusterStart;
            
            // the cluster, after which to put moved files
            endDirLcn += dirsize;
            lastDiskLcn = endDirLcn;
            
            // find files that match DirToMove & lastLcn
            AList = ALtop;
            while (AList != NULL) {
                if (CheckActivities()) break;
                
                // filter out files for moving
                if (//AList->filesize != IINVALID   && // skip dirs
                AList->Len > 0                && // skip zero byte files
                AList->fragments != LLINVALID && // skip locked
                ! AList->free                 &&
                ! AList->fragmented           && // skip fragmented
                ! AList->compressed           && // skip compressed
                ! AList->MFT                  &&
                AList->Lcn > lastDirLcn       &&
                //AList->Lcn > lastDiskLcn      &&
                _strnicmp(AList->fileName, DirToMove, lenDir) == 0 ) {
                    
                    al_size = AList->Len;
                    
                    ClusterEnd = lastDirLcn;
                    Clusters = 0;
                    while (Clusters<AList->Len && ClusterStart != LLINVALID) {
                        ClusterStart = ClusterEnd;
                        Clusters = GetFreeClusters(ClusterStart, AList->Len,
                        &ClusterStart, &ClusterEnd, TRUE);
                    }
                    
                    if (AList->filesize == IINVALID && gPARTINFO.PartitionType != 7 ) {
                        AList = AList->next;
                        continue;
                    }
                    
                    if (ClusterStart > AList->Lcn) {
                        AList = AList->next;
                        continue;
                    }
                    
                    if ( Clusters >= AList->Len   &&
                    ClusterStart < endDirLcn     ) {
                        // enough space for requested file
                        if (CheckActivities()) break;
                        
                  if (AList->filesize != IINVALID)
                         sprintf(buff, "%.3f MB - %s",
                            (float)AList->filesize/1024/1024, AList->fileName);
                      else 
                         sprintf(buff, "%s", AList->fileName);
                        UpdateStatusBar(buff, 4, 0);
                        
                        if (MoveFileNormal(AList->fileName, AList->Lcn,
                        ClusterStart, AList->Len) == AList->Len)
                        FilesMoved++;
                        AList->Lcn = ClusterStart;
                        //InsDIRFree(AList->Lcn, AList->Len);
                        //Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
                        if (lastDirLcn == AList->Lcn) {
                            lastDirLcn = AList->Lcn + AList->Len;
                            DrawBlocks(lastDirLcn, lastDirLcn+1, hpBlack, hbBlack);
                        }
                        AList = ALtop;
                        } else {
                        // move files from actual position to somewhere else
                        Here = topdirlist;
                        while (Here != NULL) {
                            if (Here->Lcn >= lastDiskLcn     &&
                            //Here->Lcn <  endDirLcn       &&
                            Here->filesize < sizeLimit   && // limit filesize in bytes
                            Here->Len > 0                && // skip zero byte files
                            Here->fragments != LLINVALID && // skip locked files
                            //Here->filesize != IINVALID   && // skip dirs
                            ! Here->fragmented           &&
                            ! Here->free                 &&
                            ! Here->compressed           && // skip compressed files
                            ! Here->MFT                  && // skip compressed files
                            _strnicmp(Here->fileName, DirToMove, lenDir) != 0 ) {
                                if (CheckActivities()) break;
                                
                                if (Here==NULL) break;
                                if (Here->filesize == IINVALID && gPARTINFO.PartitionType != 7 ) {
                                    Here = Here->next;
                                    continue;
                                }
                                
                                // from end of dirsize
                                if (GetFreeClusters(lastDiskLcn, Here->Len,
                                &ClusterStart, &ClusterEnd, TRUE) >= Here->Len) {
                           if (Here->filesize != IINVALID)
                                  sprintf(buff, "%.3f MB - %s",
                                     (float)Here->filesize/1024/1024, Here->fileName);
                               else 
                                  sprintf(buff, "%s", Here->fileName);
                                  UpdateStatusBar(buff, 4, 0);

                                    if (MoveFileNormal(Here->fileName, Here->Lcn,
                                    ClusterStart, Here->Len) == Here->Len) {
                                        FilesMoved++;
                                        cc += Here->Len;
                                        Here->Lcn = ClusterStart;
                                        //InsDIRFree(Here->Lcn, Here->Len);
                                        //Here = topdirlist = SortDict(topdirlist, listlen, byLcnDesc);
                                        if (lastDiskLcn == Here->Lcn) {
                                            lastDiskLcn = Here->Lcn + Here->Len;
                                            DrawBlocks(lastDiskLcn, lastDiskLcn+1, hpBlack, hbBlack);
                                        }
                                    }
                                }
                            }
                            if (cc >= dirsize) {
                                AList = ALtop;
                                break;
                            }
                            Here = Here->next;
                        }
                        // while for actual file in AList
                        if (CheckActivities()) break;
                        if (cc >= al_size) continue;
                    }
                    // elseif
                    
                    dirsizecounter += al_size;
                    if (dirsizecounter >= dirsize) {
                        AList = ALtop;
                        break;
                    }
                    
                }
                // if
                AList = AList->next;
            }
            // while for current DirToMove
            if (CheckActivities()) break;
            AList = ALthisdir;
            AList = AList->next;
        }
        // while for whole dirtree
        
        DelDIR(ALtop);
        
    }
    //if (CheckActivities())
    
    ghDefragFreeSpace = NULL;
    if (ghDriveChange != NULL) {
        FindCloseChangeNotification(ghDriveChange);
        ghDriveChange = NULL;
    }
    
    // re-enable menu
    wsprintf(buff, "Defrag by sor&ting Directories\tF10");
    hsub = GetSubMenu(GetMenu(ghWnd), 0);
    DeleteMenu(hsub, 5, MF_BYPOSITION);
    InsertMenu(hsub, 5, MF_STRING|MF_BYPOSITION|MF_ENABLED, IDM_DIRECTORYDEFRAG, buff);
    EnableMenuItem(hsub, 0, MF_BYPOSITION|MF_ENABLED);
    EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
    EnableMenuItem(hsub, 2, MF_BYPOSITION|MF_ENABLED);
    EnableMenuItem(hsub, 3, MF_BYPOSITION|MF_ENABLED);
    hsub = GetMenu(ghWnd);
    EnableMenuItem(hsub, 1, MF_BYPOSITION|MF_ENABLED);
    DrawMenuBar(ghWnd);
    
    wsprintf(buff, "");
    if (FilesMoved>0) {
        wsprintf(buff, "%s file(s) moved", fmtNumber(FilesMoved, buff));
        ResetStatusBar();
        
        bAnalyzed = FALSE;
        // draw disk-layout
        DrawBitMap(NULL);
        // do an new Analyze
        //SendMessage(ghWnd, WM_COMMAND, IDM_ANALYZE, 0);
    }
    
    //OutLine2(buff);
    OutLine3(buff);
    UpdateStatusBar("DefragBySortingDirs finished", 4, 0);
}
