#include <windows.h>
#include <wingdi.h>
#include <winioctl.h> 
#include <commctrl.h> 

// Invalid values
#define LLINVALID   ((ULONGLONG) -1)
#define IINVALID    (  (__int64) -1)

#define SZDESCRIPTION "simple Defragger"

// USER_MESSAGES
#define WM_DRIVECHANGED      6666
#define WM_UPDATECLUSTERLIST 6667
#define WM_UPDATEDRIVEINFO   6668

#define MM_DRIVE_NUM    8301 // ..8326, one for each drive letter

#define DRIVE_MENU_NUM      1      // This is the 'Drives' submenu

#define NUM_POSSIBLE_REMOVABLE_DRIVES   26      // A-Z

#define MAX_LOADSTRING 100
#define PAINTCURSOR    1002

// This is the definition for the data structure that is passed in to
// FSCTL_GET_NTFS_VOLUME_DATA
typedef struct {
    LARGE_INTEGER VolumeSerialNumber;
    LARGE_INTEGER NumberSectors;
    LARGE_INTEGER TotalClusters;
    LARGE_INTEGER FreeClusters;
    LARGE_INTEGER TotalReserved;
    DWORD BytesPerSector;
    DWORD BytesPerCluster;
    DWORD BytesPerFileRecordSegment;
    DWORD ClustersPerFileRecordSegment;
    LARGE_INTEGER MftValidDataLength;
    LARGE_INTEGER MftStartLcn;
    LARGE_INTEGER Mft2StartLcn;
    LARGE_INTEGER MftZoneStart;
    LARGE_INTEGER MftZoneEnd;
} NTFS_DATA, *PNTFS_DATA;

typedef struct {
  LARGE_INTEGER FileReferenceNumber;
} NTFS_FILE_RECORD_INPUT_BUFFER, *PNTFS_FILE_RECORD_INPUT_BUFFER;

typedef struct {
  LARGE_INTEGER FileReferenceNumber;
  DWORD FileRecordLength;
  BYTE FileRecordBuffer[1];
} NTFS_FILE_RECORD_OUTPUT_BUFFER, *PNTFS_FILE_RECORD_OUTPUT_BUFFER;

struct {
	ULONGLONG start;
	ULONGLONG end;
} MFTZONE[3];

// global Handles
HINSTANCE ghInst;         //act Instance
HWND ghWnd;               //for window
HWND ghStatWnd;           //statusbar
HWND ghListWnd;           //Handle to list view window

HANDLE ghVolume;          //for the raw volume that was opened
HANDLE ghDraw;            //analyze thread
HANDLE ghAnalyze;         //analyze thread
HANDLE ghDefragFragments; //defragfragments thread
HANDLE ghDefragFreeSpace; //defragfreespace thread
HANDLE ghDefrag;          //defrag thread
HANDLE ghDriveChange;     //drive listener
FILE_NOTIFY_INFORMATION    ghFNI;     //result of drive listener

PARTITION_INFORMATION      gPARTINFO;
DISK_GEOMETRY              gDISKGEOM;
NTFS_DATA                  gNTFSDATA;
NTFS_FILE_RECORD_INPUT_BUFFER gNTFSRECORDi;
NTFS_FILE_RECORD_OUTPUT_BUFFER gNTFSRECORDo;

// global Variables
BOOL   bAnalyzed;
BOOL   bDriveChanged;
BOOL   bStopAction;
BOOL   bShowMarks;

// global char vars
char cDriveLetters[26];          // hold DriveLetters
char cActualDrive;
char cVolName[32], cFileSystem[16], cWinDir[MAX_PATH], cSysDir[MAX_PATH], cSysDir2[MAX_PATH];
char LastOutLine2[MAX_PATH], LastOutLine3[MAX_PATH];

// essential draw variables
int           imgWidth, imgHeight;
int           blocksize, blockcount, bw, bh;
int           dist; // distance of graphic from statusbar
int           xposlast, yposlast; // position of last outlined block
__int64       sizeLimit;

// counter vars start with c
int           clustersPB;
ULONGLONG     cDiskClusters, cFreeClusters, clustersPL;
__int64       cBytes, cBytesSystem, cBytesFree, cBytesUsed, cBytesTotal;
unsigned long cRootDirs, cDirs, cFiles, cDirsFiles;
unsigned long cFragmented, cLocked, cCompressed, cFragmentsAllTogether;

// functions called as threads
void          LockUpDriveChange(void *dummy);
void          DrawBitMap(void *dummy);
void          AnalyzeVolume(void *dummy);
void          DefragFragments(void *dummy);
void          DefragFreeSpace(void *dummy);
void          DefragByDirectories(void *dummy);
void          Defrag(void *dummy);

// functions called normal
void          GetDiskGeometry(void);
void          GetDiskParameters(void);
void          UpdateClusterList(void);
void          UpdateDriveInformation(void);
void          DrawNoBlockBound(int, int);
void          DrawBlockBound(int, int);
void          ResetStatusBar(void);
char          *fmtNumber(__int64 _in, char *cGes);
int           f_instr(const char *s1, const char *s2);
BOOL          CheckActivities(void);

// OutLine0-3: TextOut at line 0-3 over the statusbar
void OutLine0(char *buff);
void OutLine1(char *buff);
void OutLine2(char *buff);
void OutLine3(char *buff);

// graphic stuff
HDC         memDC;        // memory device context
HBITMAP     memBM, oldBM; // memory bitmap (for  memDC) Double buffering data
                          //    old bitmap (from memDC)
HICON hiDir, hiFile, hiFrag, hiComp, hiMFT, hiFree, hiSystem, hiMain;

COLORREF ColBG;
COLORREF ColDGreen  , ColGreen   , ColLGreen  , ColL2Green , ColL3Green , ColVio     , ColRed     , ColOck  ;
COLORREF ColYellow  , ColLYellow , ColBlue    , ColLBlue   , ColL2Blue  , ColBrown;
COLORREF ColWhite   , ColGray    , ColBlack   ;

HBRUSH      hbBG;
HPEN        hpBG;

HPEN        hpDGreen, hpGreen, hpLGreen, hpL2Green, hpL3Green, hpVio, hpRed, hpOck;
HPEN        hpYellow, hpLYellow, hpBlue, hpLBlue, hpL2Blue, hpBrown;
HPEN        hpWhite, hpGray, hpBlack;

HBRUSH      hbDGreen, hbGreen, hbLGreen, hbL2Green, hbL3Green, hbVio, hbRed, hbOck, hbYellow, hbLYellow;
HBRUSH      hbBlue, hbLBlue, hbL2Blue, hbBrown, hbWhite, hbGray, hbBlack;
