//list.c

#include "globals.h"
#include "list.h"

#define Compare(a,b) ( strcmp(a,b) )  // used by SortDict

struct DICT *InsDIR(char *name, __int64 size) {
    struct DICT *newel;

    // neues DirTree-Element anlegen
    newel = (struct DICT*) malloc (sizeof(struct DICT));
    newel->fileName = (char *)malloc(strlen(name) + 1);
    strcpy(newel->fileName, name);
    newel->filesize = size;
    newel->fragments  = 0;
    newel->free       = FALSE;
    newel->fragmented = FALSE;
    newel->compressed = FALSE;
    newel->MFT        = FALSE;
    if (lstrcmp(name, "") == 0 && size == 0) 
        newel->free = TRUE;

    if (dirlist==NULL) {            //liste da ?
        newel->next   = dirlist;    //wenn nein, dann anlegen
        topdirlist    = dirlist = newel;
    } else {
        dirlist->next = newel;      // ansonsten an ende anfuegen
        newel->next   = NULL;
        dirlist       = dirlist->next;
    }
    listlen++;
    return newel;
}

struct DICT *CopyDIR(struct DICT *top) {
    struct DICT *work, *newel, *newlist=NULL, *topnewlist;

    if (top==NULL) return NULL;
    topnewlist=newel=newlist=NULL;
    work=top;                 /* alles an den Listenanfang */
    while (work != NULL) {
        // neues DirTree-Element anlegen
        newel             = (struct DICT*) malloc (sizeof(struct DICT));
        if (newel == NULL) return newlist;
        newel->fileName   = (char *)malloc(strlen(work->fileName) + 1);
        if (newel->fileName == NULL) return topnewlist;
        strcpy(newel->fileName, work->fileName);
        newel->Len    = work->Len;
        newel->Lcn        = work->Lcn;
        newel->filesize   = work->filesize;
        newel->free       = work->free;
        newel->compressed = work->compressed;
        newel->fragmented = work->fragmented;
        newel->fragments  = work->fragments;
        newel->MFT        = work->MFT;

        if (newlist == NULL) {          //erstes schon angelegt?
            newel->next   = newlist;    //wenn nein, dann anlegen
            newlist       = newel;
            topnewlist    = newlist;
        } else {
            newlist->next = newel;      // ansonsten an ende anfuegen
            newel->next   = NULL;
            newlist       = newlist->next;
        }
        work = work->next;
    }
    return topnewlist;
}

void DelDIR(struct DICT *top) {
    struct DICT *last, *work;
    if (top == NULL) return;

    last=work=top;                 /* alles an den Listenanfang */
    while (work!=NULL) {
        if (work->next==NULL) {    /* am Ende der Liste */
            if (work!=top) {       /* und nicht am Anfang */
                last->next=NULL;   /* Listenende festlegen */
                free(work->fileName);
                free(work);        /* altes El. freigeben */
                work=NULL;
            } else {              /* ansonsten hat die Liste nur ein */
                top=NULL;         /* Element */
                free(work->fileName);
                free(work);
                work=NULL;
            }
        } else {
            if (work==top) {      /* am Anfang? */
                top=work->next;   /* top neu setzen */
                free(work->fileName);
                free(work);
                work=top;
            } else {
                last->next=work->next;/* Zeiger an zu loeschenden vorbeizeigen lassen */
                free(work->fileName);
                free(work);      /* und dieses dann loeschen */
                work=last;
            }
        }
    } // while 
}

///////////////////////////////////////////////////////////////////////
// this sorting routine is robbed - dont know where 
// thanks to the author
///////////////////////////////////////////////////////////////////////
/* Sort a list of struct DICT by using an iterative bottom-up merge sort.
 * This particular piece of code took me ages to do (well, 2 days + 3 weeks
 * research) and provides a FAST way of sorting a linked list without the
 * overhead of increasing memory usage via malloc() or brk(). Why ? Because I
 * have to assume that there is no more memory, thats why. It's all Brian
 * Thompsett's fault! Really! Filling the swapspace on a SparcStation2 and
 * expecting Crack to survive! Argh! 8-) */

// Since this code is so nice, I'll comment it fairly thoroughly //
struct DICT * SortDict (register struct DICT *chain3, 
                        long int listlength, 
                        int sortItem)
{
    // misc counters //
    register int i;
    long int discarded;

    long int n; // 2^n for n = 0..x //

    register struct DICT *chain1;  // head of the first extracted subchain //
    register struct DICT *chain2;  // head of second subchain //
    register struct DICT *scratch; // useful temp pointer //
    struct DICT *lead_in;          // PTR TO ELEMENT containing TAIL of unsorted list pre-merging //
    struct DICT *lead_out;         // PTR TO HEAD of unsorted list after extracting chains //
    struct DICT dummy1;            // dummy structures used as fenceposts //
    struct DICT dummy2;

    // Put the incoming list into 'dummy1' posthole //
    dummy1.next = chain3;

    // For values of n = 2^(0..30) limited by listlength //
    for (n = 1L; n < listlength; n *= 2) {
      // Store place to get/put head of list in 'lead_in' //
      lead_in = &dummy1;

      // Set chain1 to the head of unsorted list //
      for (chain1 = lead_in->next; chain1; chain1 = lead_in->next) {
         // Break connection head and chain1 //
         lead_in->next = (struct DICT *) 0;

         // Extract up to length 'n', park on last element before chain2 //
         for (i = n - 1, scratch = chain1;
             i && scratch->next;
             scratch = scratch->next) {
            i--;
         };

         // If chain1 is undersized/exact, there is no chain2 //
         if (i || !scratch->next) {
            // put chain1 back where you got it and break //
            lead_in->next = chain1;
            break;
         }

         // Get pointer to head of chain2 //
         chain2        = scratch->next;

         // Break connection between chain1 & chain2 //
         scratch->next = (struct DICT *) 0;

         // Extract up to length 'n', park on last element of chain2 //
         for (i = n - 1, scratch = chain2;
             i && scratch->next;
             scratch = scratch->next) {
            i--;
         };

         // Even if it's NULL, store rest of list in 'lead_out' //
         lead_out      = scratch->next;

         // Break connection between chain2 & tail of unsorted list //
         scratch->next = (struct DICT *) 0;

         // Now, mergesort chain1 & chain2 to chain3 //

         // Set up dummy list fencepost //
         chain3       = &dummy2;
         chain3->next = (struct DICT *) 0;

         // While there is something in each list //
         while (chain1 && chain2) {
            switch (sortItem) {
            case 1 :
               // ---byname
               i = Compare (chain1->fileName, chain2->fileName);
               if (i > 0) {
                  // a < b
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
               } else if (i < 0) {
                  // a > b
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               } else {
                  // a == b. Link them both in. Don't try to get rid of the
                  // multiple copies here, because if you free up any
                  // elements at this point the listsize changes and the
                  // algorithm runs amok.
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               }
               break;
            case 2 :
               // ---byLcn aufsteigend
               if (chain1->Lcn < chain2->Lcn) {
                  // a < b
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
               } else if (chain1->Lcn > chain2->Lcn) {
                  // a > b
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               } else {
                  // a == b. Link them both in. Don't try to get rid of the
                  // multiple copies here, because if you free up any
                  // elements at this point the listsize changes and the
                  // algorithm runs amok.
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               }
               break;

            case 3 :
               // ---byLcn abfsteigend
               if (chain1->Lcn > chain2->Lcn) {
                  // a < b
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
               } else if (chain1->Lcn < chain2->Lcn) {
                  // a > b
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               } else {
                  // a == b. Link them both in. Don't try to get rid of the
                  // multiple copies here, because if you free up any
                  // elements at this point the listsize changes and the
                  // algorithm runs amok.
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               }
               break;

            case 4 :
               // ---byfragments
               if (chain1->fragments > chain2->fragments) {
                  // a < b
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
               } else if (chain1->fragments < chain2->fragments) {
                  // a > b
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               } else {
                  // a == b. Link them both in. Don't try to get rid of the
                  // multiple copies here, because if you free up any
                  // elements at this point the listsize changes and the
                  // algorithm runs amok.
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               }
               break;
            case 5 :
               // ---bySizeAsc
               if (chain1->Len < chain2->Len) {
                  // a < b
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
               } else if (chain1->Len > chain2->Len) {
                  // a > b
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               } else {
                  // a == b. Link them both in. Don't try to get rid of the
                  // multiple copies here, because if you free up any
                  // elements at this point the listsize changes and the
                  // algorithm runs amok.
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               }
               break;
            case 6 :
               // ---bySizeAsc
               if (chain1->Len > chain2->Len) {
                  // a < b
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
               } else if (chain1->Len < chain2->Len) {
                  // a > b
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               } else {
                  // a == b. Link them both in. Don't try to get rid of the
                  // multiple copies here, because if you free up any
                  // elements at this point the listsize changes and the
                  // algorithm runs amok.
                  chain3->next = chain1;
                  chain3       = chain1;
                  chain1       = chain1->next;
                  chain3->next = chain2;
                  chain3       = chain2;
                  chain2       = chain2->next;
               }
               break;

            } // switch
         } // while

         // Whatever is left is sorted and therefore linkable straight
         // onto the end of the current list
         if (chain1) {
            chain3->next = chain1;
         } else {
            chain3->next = chain2;
         }

         // Skip to the end of the sorted list //
         while (chain3->next) chain3 = chain3->next;

         // Append this lot to where you got chain1 from ('lead_in') //
         lead_in->next = dummy2.next;

         // Append rest of unsorted list to chain3 //
         chain3->next = lead_out;

         // Set 'lead_in' for next time to last element of 'chain3' //
         lead_in = chain3;
      } // for 2
    } // for 1

    // Now, Uniq the list //
    discarded = 0;

    // Chain1 to the head of the list, Chain2 to the next
    chain1 = dummy1.next;
    chain2 = chain1->next;

    // While not at end of list
    while (chain2) {
      // Whilst (chain1) == (chain2)
      while (!Compare (chain1->fileName, chain2->fileName) &&
             chain1->Lcn == chain2->Lcn &&
             chain1->filesize == chain2->filesize &&
             chain1->fragments == chain2->fragments
            ) {
         // Bump the discard count 
         discarded++;

         // Store the next element 
         scratch = chain2->next;

         // Get some memory back 
         free (chain2);   // ...<snigger>... 

         // Assign the skip, break if you run off the end of list 
         if (!(chain2 = scratch)) break;
      }

      // Set comparison ptr to new element or terminate 
      chain1->next = chain2;

      // If not terminated 
      if (chain2) {
         // set the compared pointer to its successor 
         chain1 = chain2;
         chain2 = chain2->next;
      }
    }

    {
        //char buff[255];
        //wsprintf(buff, "Sort discarded %ld words, dict-size now: %ld",
        //               discarded, listlength - discarded);
        //OutLine1(buff);
    }
    return (dummy1.next);
}
